# Copyright (c) OpenMMLab. All rights reserved.
import os
from pathlib import Path
from typing import Dict, List, Optional, Union

import torch
from mmengine.model import BaseModel, MMDistributedDataParallel
from mmengine.optim import OptimWrapper
from mmengine.structures import BaseDataElement
from torch import nn

from mmrazor.models.distillers import ConfigurableDistiller
from mmrazor.models.mutators import OneShotChannelMutator
from mmrazor.models.utils import (add_prefix,
                                  reinitialize_optim_wrapper_count_status)
from mmrazor.registry import MODEL_WRAPPERS, MODELS
from mmrazor.utils import SingleMutatorRandomSubnet
from ..base import BaseAlgorithm

VALID_MUTATOR_TYPE = Union[OneShotChannelMutator, Dict]
VALID_DISTILLER_TYPE = Union[ConfigurableDistiller, Dict]
VALID_PATH_TYPE = Union[str, Path]
VALID_CHANNEL_CFG_PATH_TYPE = Union[VALID_PATH_TYPE, List[VALID_PATH_TYPE]]


@MODELS.register_module()
class AutoSlim(BaseAlgorithm):

    def __init__(self,
                 mutator: VALID_MUTATOR_TYPE,
                 distiller: VALID_DISTILLER_TYPE,
                 architecture: Union[BaseModel, Dict],
                 data_preprocessor: Optional[Union[Dict, nn.Module]] = None,
                 init_cfg: Optional[Dict] = None,
                 num_samples: int = 2) -> None:
        super().__init__(architecture, data_preprocessor, init_cfg)

        self.mutator = self._build_mutator(mutator)
        # `prepare_from_supernet` must be called before distiller initialized
        self.mutator.prepare_from_supernet(self.architecture)

        self.distiller = self._build_distiller(distiller)
        self.distiller.prepare_from_teacher(self.architecture)
        self.distiller.prepare_from_student(self.architecture)

        self.num_samples = num_samples

        self._optim_wrapper_count_status_reinitialized = False

    def _build_mutator(self,
                       mutator: VALID_MUTATOR_TYPE) -> OneShotChannelMutator:
        """build mutator."""
        if isinstance(mutator, dict):
            mutator = MODELS.build(mutator)
        if not isinstance(mutator, OneShotChannelMutator):
            raise TypeError('mutator should be a `dict` or '
                            '`OneShotModuleMutator` instance, but got '
                            f'{type(mutator)}')

        return mutator

    def _build_distiller(
            self, distiller: VALID_DISTILLER_TYPE) -> ConfigurableDistiller:
        if isinstance(distiller, dict):
            distiller = MODELS.build(distiller)
        if not isinstance(distiller, ConfigurableDistiller):
            raise TypeError('distiller should be a `dict` or '
                            '`ConfigurableDistiller` instance, but got '
                            f'{type(distiller)}')

        return distiller

    def sample_subnet(self) -> SingleMutatorRandomSubnet:
        return self.mutator.sample_choices()

    def set_subnet(self, subnet: SingleMutatorRandomSubnet) -> None:
        self.mutator.set_choices(subnet)

    def set_max_subnet(self) -> None:
        self.mutator.set_max_choices()

    def set_min_subnet(self) -> None:
        return self.mutator.set_min_choices()

    def train_step(self, data: List[dict],
                   optim_wrapper: OptimWrapper) -> Dict[str, torch.Tensor]:

        def distill_step(
                batch_inputs: torch.Tensor, data_samples: List[BaseDataElement]
        ) -> Dict[str, torch.Tensor]:
            subnet_losses = dict()
            with optim_wrapper.optim_context(
                    self), self.distiller.student_recorders:  # type: ignore
                hard_loss = self(batch_inputs, data_samples, mode='loss')
                soft_loss = self.distiller.compute_distill_losses()

                subnet_losses.update(hard_loss)
                subnet_losses.update(soft_loss)

                parsed_subnet_losses, _ = self.parse_losses(subnet_losses)
                optim_wrapper.update_params(parsed_subnet_losses)

            return subnet_losses

        if not self._optim_wrapper_count_status_reinitialized:
            reinitialize_optim_wrapper_count_status(
                model=self,
                optim_wrapper=optim_wrapper,
                accumulative_counts=self.num_samples + 2)
            self._optim_wrapper_count_status_reinitialized = True

        batch_inputs, data_samples = self.data_preprocessor(data, True)

        total_losses = dict()
        self.set_max_subnet()
        with optim_wrapper.optim_context(
                self), self.distiller.teacher_recorders:  # type: ignore
            max_subnet_losses = self(batch_inputs, data_samples, mode='loss')
            parsed_max_subnet_losses, _ = self.parse_losses(max_subnet_losses)
            optim_wrapper.update_params(parsed_max_subnet_losses)
        total_losses.update(add_prefix(max_subnet_losses, 'max_subnet'))

        self.set_min_subnet()
        min_subnet_losses = distill_step(batch_inputs, data_samples)
        total_losses.update(add_prefix(min_subnet_losses, 'min_subnet'))

        for sample_idx in range(self.num_samples):
            self.set_subnet(self.sample_subnet())
            random_subnet_losses = distill_step(batch_inputs, data_samples)
            total_losses.update(
                add_prefix(random_subnet_losses,
                           f'random_subnet_{sample_idx}'))

        return total_losses


@MODEL_WRAPPERS.register_module()
class AutoSlimDDP(MMDistributedDataParallel):

    def __init__(self,
                 *,
                 device_ids: Optional[Union[List, int, torch.device]] = None,
                 **kwargs) -> None:
        if device_ids is None:
            if os.environ.get('LOCAL_RANK') is not None:
                device_ids = [int(os.environ['LOCAL_RANK'])]
        super().__init__(device_ids=device_ids, **kwargs)

    def train_step(self, data: List[dict],
                   optim_wrapper: OptimWrapper) -> Dict[str, torch.Tensor]:

        def distill_step(
                batch_inputs: torch.Tensor, data_samples: List[BaseDataElement]
        ) -> Dict[str, torch.Tensor]:
            subnet_losses = dict()
            with optim_wrapper.optim_context(
                    self
            ), self.module.distiller.student_recorders:  # type: ignore
                hard_loss = self(batch_inputs, data_samples, mode='loss')
                soft_loss = self.module.distiller.compute_distill_losses()

                subnet_losses.update(hard_loss)
                subnet_losses.update(soft_loss)

                parsed_subnet_losses, _ = self.module.parse_losses(
                    subnet_losses)
                optim_wrapper.update_params(parsed_subnet_losses)

            return subnet_losses

        if not self._optim_wrapper_count_status_reinitialized:
            reinitialize_optim_wrapper_count_status(
                model=self,
                optim_wrapper=optim_wrapper,
                accumulative_counts=self.module.num_samples + 2)
            self._optim_wrapper_count_status_reinitialized = True

        batch_inputs, data_samples = self.module.data_preprocessor(data, True)

        total_losses = dict()
        self.module.set_max_subnet()
        with optim_wrapper.optim_context(
                self), self.module.distiller.teacher_recorders:  # type: ignore
            max_subnet_losses = self(batch_inputs, data_samples, mode='loss')
            parsed_max_subnet_losses, _ = self.module.parse_losses(
                max_subnet_losses)
            optim_wrapper.update_params(parsed_max_subnet_losses)
        total_losses.update(add_prefix(max_subnet_losses, 'max_subnet'))

        self.module.set_min_subnet()
        min_subnet_losses = distill_step(batch_inputs, data_samples)
        total_losses.update(add_prefix(min_subnet_losses, 'min_subnet'))

        for sample_idx in range(self.module.num_samples):
            self.module.set_subnet(self.module.sample_subnet())
            random_subnet_losses = distill_step(batch_inputs, data_samples)
            total_losses.update(
                add_prefix(random_subnet_losses,
                           f'random_subnet_{sample_idx}'))

        return total_losses

    @property
    def _optim_wrapper_count_status_reinitialized(self) -> bool:
        return self.module._optim_wrapper_count_status_reinitialized

    @_optim_wrapper_count_status_reinitialized.setter
    def _optim_wrapper_count_status_reinitialized(self, val: bool) -> None:
        assert isinstance(val, bool)

        self.module._optim_wrapper_count_status_reinitialized = val



# dataset settings
dataset_type = 'mmcls.ImageNet'
preprocess_cfg = dict(
    # RGB format normalization parameters
    mean=[123.675, 116.28, 103.53],
    std=[58.395, 57.12, 57.375],
    # convert image from BGR to RGB
    to_rgb=True,
)

train_pipeline = [
    dict(type='mmcls.LoadImageFromFile'),
    dict(type='mmcls.RandomResizedCrop', scale=224),
    dict(
        type='mmcls.ColorJitter', brightness=0.4, contrast=0.4,
        saturation=0.4),
    dict(type='mmcls.RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='mmcls.PackClsInputs'),
]

test_pipeline = [
    dict(type='mmcls.LoadImageFromFile'),
    dict(type='mmcls.ResizeEdge', scale=256, edge='short'),
    dict(type='mmcls.CenterCrop', crop_size=224),
    dict(type='mmcls.PackClsInputs'),
]

train_dataloader = dict(
    batch_size=128,
    num_workers=4,
    dataset=dict(
        type=dataset_type,
        data_root='data/imagenet',
        ann_file='meta/train.txt',
        data_prefix='train',
        pipeline=train_pipeline),
    sampler=dict(type='mmcls.DefaultSampler', shuffle=True),
    persistent_workers=True,
)

val_dataloader = dict(
    batch_size=128,
    num_workers=4,
    dataset=dict(
        type=dataset_type,
        data_root='data/imagenet',
        ann_file='meta/val.txt',
        data_prefix='val',
        pipeline=test_pipeline),
    sampler=dict(type='mmcls.DefaultSampler', shuffle=False),
    persistent_workers=True,
)
val_evaluator = dict(type='mmcls.Accuracy', topk=(1, 5))

# If you want standard test, please manually configure the test dataset
test_dataloader = val_dataloader
test_evaluator = val_evaluator

# optimizer
paramwise_cfg = dict(
    bias_decay_mult=0.0, norm_decay_mult=0.0, dwconv_decay_mult=0.0)

optim_wrapper = dict(
    optimizer=dict(type='mmcls.SGD', lr=0.5, momentum=0.9, weight_decay=4e-5),
    paramwise_cfg=paramwise_cfg,
    clip_grad=None)

# leanring policy
param_scheduler = dict(
    type='mmcls.PolyLR',
    power=1.0,
    eta_min=0.0,
    by_epoch=True,
    end=300,
    convert_to_iter_based=True)

# train, val, test setting
train_cfg = dict(by_epoch=True, max_epochs=300)
val_cfg = dict()
test_cfg = dict()



# Copyright (c) OpenMMLab. All rights reserved.
"""This module defines ModuleNode and ModuleGraph.

They model the computation graph of a model based on BaseNode and BaseGraph
"""
import copy
from collections import OrderedDict
from typing import Dict, List, TypeVar, Union

import torch.nn as nn
from torch.nn import Module

from mmrazor.models.task_modules import (BackwardTracer, Path, PathConcatNode,
                                         PathList, PathNode)
from mmrazor.models.task_modules.tracer import ImageClassifierPseudoLoss
from .base_graph import BaseGraph, BaseNode


class ModuleNode(BaseNode):
    """A node in a computation graph.

    All nodes are divided to four types, the detail of definition can be found
    in functions  self.is_{xxx}_node.
    """

    pre_defined_node_val_str = [
        'cat_placeholder', 'bind_placeholder', 'pass_placeholder'
    ]

    def __init__(self,
                 name: str,
                 val: Union[Module, str],
                 expand_ratio: int = 1) -> None:
        """
        Args:
            name (str): the name of the node
            val (Module | str): content of the node. It can be Module or
            string. If val is a string, the string can only be one of
                self.pre_defined_node_val_str
            expand_ratio (int): expand_ratio is used in bind node,
                where the out_channel is always a multiple of the in_channel.
        Note:
            Here, we give an example of expand_ratio.
            >>> class Pool(nn.Module):
                    def forward(x):
                        return F.adaptive_avg_pool2d(x,2).flatten(1)
            >>> node= ModuleNode('pass_0',Pool(),expand_ratio=4)
            >>> assert node.out_channels == node.in_channels*4
        """

        assert (isinstance(val, Module)
                or val in self.__class__.pre_defined_node_val_str
                ), f'{val} node is not allowed'
        if expand_ratio != 1:
            assert val == 'pass_placeholder', \
                'expand != 1 is only valid when val=="pass"'
        super().__init__(name, val)
        self.expand_ratio = expand_ratio

    # channel

    @property
    def in_channels(self) -> int:
        """int: the in_channels of the node."""
        if isinstance(self.val, nn.Module):
            MAPPING = {
                nn.Conv2d: 'in_channels',
                nn.modules.batchnorm._BatchNorm: 'num_features',
                nn.modules.Linear: 'in_features',
            }
            for basetype in MAPPING:
                if isinstance(self.val, basetype):
                    return getattr(self.val, MAPPING[basetype])
            raise NotImplementedError(f'unsupported module: {self.val}')
        elif self.is_bind_node() or self.is_pass_node():
            if len(self.prev_nodes) > 0:
                return self.prev_nodes[0].out_channels
            else:
                return 0
        elif self.is_cat_node():
            return sum([
                node.out_channels if node.out_channels is not None else 0
                for node in self.prev_nodes
            ])
        else:
            raise NotImplementedError(
                f'unsupported node type: {self.basic_type}')

    @property
    def out_channels(self) -> int:
        """int: the out_channels of the node."""
        if isinstance(self.val, nn.Module):
            MAPPING = {
                nn.Conv2d: 'out_channels',
                nn.modules.batchnorm._BatchNorm: 'num_features',
                nn.modules.Linear: 'out_features',
            }
            for basetype in MAPPING:
                if isinstance(self.val, basetype):
                    return getattr(self.val, MAPPING[basetype])
            raise NotImplementedError(f'unsupported module: {self.val}')
        elif self.is_bind_node():
            if len(self.prev_nodes) > 0:
                return self.prev_nodes[0].out_channels
            else:
                return 0
        elif self.is_pass_node():
            return self.in_channels * self.expand_ratio
        elif self.is_cat_node():
            return sum([
                node.out_channels if node.out_channels is not None else 0
                for node in self.prev_nodes
            ])
        else:
            raise NotImplementedError(
                f'unsupported node type: {self.basic_type}')

    # other

    def __repr__(self) -> str:
        return f'{self.name}_({self.in_channels},{self.out_channels})'

    # node type

    @property
    def basic_type(self) -> str:
        """The basic type of the node.

        Basic types are divided into seveval major types, detailed in
        self.is_{xxx}_node
        """
        if isinstance(self.val, Module):
            if isinstance(self.val, nn.Conv2d):
                if self.val.groups == 1:
                    return 'conv2d'
                elif self.val.groups == self.val.in_channels == \
                        self.val.out_channels:
                    return 'dwconv2d'
                else:
                    return 'gwconv2d'
            elif isinstance(self.val, nn.modules.batchnorm._BatchNorm):
                return 'bn'
            elif isinstance(self.val, nn.Linear):
                return 'linear'
            else:
                raise NotImplementedError(f'{self}')
        else:
            if self.val in [
                    'cat_placeholder', 'bind_placeholder', 'pass_placeholder'
            ]:
                return self.val
            else:
                raise NotImplementedError()

    def is_pass_node(self):
        """pass node represent a module whose in-channels correspond out-
        channels one-to-one."""
        return self.basic_type in ['bn', 'dwconv2d', 'pass_placeholder']

    def is_cat_node(self):
        """cat node represents a cat module."""
        return self.basic_type == 'cat_placeholder'

    def is_bind_node(self):
        """bind node represent a node that has multiple inputs, and their
        channels are bound one-to-one."""
        return self.basic_type == 'bind_placeholder'

    def is_mix_node(self):
        """mix node represents a module that mixs all input channels and
        generete new output channels, such as conv and linear."""
        return self.basic_type in ['conv2d', 'linear', 'gwconv2d']

    # check

    def check_channel(self):
        """Check if the channels of the node is matchable with previous nodes
        and next nodes."""
        if self.is_cat_node():
            pass
        else:
            for pre in self.prev_nodes:
                assert pre.out_channels == self.in_channels, \
                    f'{self} has channel error'

    def check_type(self):
        """Check if the node has right number of previous nodes according to
        their type."""
        if self.is_pass_node():
            assert len(self.prev_nodes) <= 1, '{name} pass node error'
        elif self.is_cat_node():
            pass
        elif self.is_bind_node():
            assert len(self.prev_nodes) > 1, '{name} bind node error'
        elif self.is_mix_node():
            assert len(self.prev_nodes) <= 1, '{name} mix node error'
        else:
            raise NotImplementedError(f'{self}')


MODULENODE = TypeVar('MODULENODE', bound=ModuleNode)


class ModuleGraph(BaseGraph[MODULENODE]):
    """Computatation Graph."""

    def __init__(self) -> None:
        super().__init__()
        self._model = None

    # functions to generate module graph.

    @staticmethod
    def init_using_backward_tracer(
        model: Module,
        backward_tracer=BackwardTracer(
            loss_calculator=ImageClassifierPseudoLoss()),
    ):
        """init module graph using backward tracer."""
        path_lists = backward_tracer.trace(model)
        converter = PathToGraphConverter(path_lists, model)
        return converter.graph

    @staticmethod
    def init_using_fx_tracer(model: Module, is_extra_leaf_module=None):
        """init module graph using torch fx tracer."""
        pass

    @staticmethod
    def init_from_model(model: Module):
        """init module graph from a model which uses connect_module to record
        the relation among modules."""
        pass

    # check

    def check(self):
        """Check if the graph is valid."""
        for node in self:
            node.check_channel()
            node.check_type()

    # static method for models that can't use tracer

    @staticmethod
    def connect_module(pre: Module, next: Module):
        """This function is used to write hardcode in modules to generate Graph
        object using init_from_model."""
        if hasattr(pre, '_next'):
            _next = getattr(pre, '_next')
            assert isinstance(_next, List)
        else:
            pre._next = set()
        pre._next.add(next)

        if hasattr(next, '_pre'):
            _pre = getattr(next, '_pre')
            assert isinstance(_pre, List)
        else:
            next._pre = set()
        next._pre.add(pre)


# Converter


class GraphConverter:
    """Base class for converters for ModuleGraph."""

    def __init__(self) -> None:
        self.graph = ModuleGraph[ModuleNode]()
        self.cat_placeholder_num = 0
        self.bind_placeholder_num = 0
        self.pass_placeholder_num = 0

    # add node

    def _new_placeholder_node(self, type: str, expand_ratio=1):
        """New cat/bind/pass node."""
        assert type in [
            'cat_placeholder', 'pass_placeholder', 'bind_placeholder'
        ]
        if expand_ratio != 1:
            assert type == 'pass_placeholder'
        if type == 'cat_placeholder':
            num = self.cat_placeholder_num
            self.cat_placeholder_num += 1
        elif type == 'pass_placeholder':
            num = self.pass_placeholder_num
            self.pass_placeholder_num += 1
        elif type == 'bind_placeholder':
            num = self.bind_placeholder_num
            self.bind_placeholder_num += 1
        else:
            pass
        node = ModuleNode(f'{type}_{num}', type, expand_ratio=expand_ratio)
        self.graph.add_or_find_node(node)
        return node

    # insert nodes

    def _insert_node_before(self, node: ModuleNode, new_node: ModuleNode):
        """Insert a new node before a node."""
        for pre in node.prev_nodes:
            self.graph.connect(pre, new_node)
        for pre in new_node.prev_nodes:
            self.graph.disconnect(pre, node)
        self.graph.connect(new_node, node)

    def _insert_bind_nodes(self):
        """Add bind nodes before the nodes which only need one previous node
        but have more than one."""

        need_bind_nodes = []
        for node in self.graph:
            if (isinstance(node.val, nn.Conv2d)
                    or isinstance(node.val, nn.Linear)
                    or isinstance(node.val, nn.modules.batchnorm._BatchNorm)):
                if len(node.prev_nodes) > 1:
                    need_bind_nodes.append(node)
        for node in need_bind_nodes:
            bind_node = self._new_placeholder_node('bind_placeholder')
            self._insert_node_before(node, bind_node)

    def _insert_pass_nodes(self):
        """Add pass nodes where the channel conflict."""
        for node in copy.copy(list(self.graph.nodes.values())):
            if len(node.prev_nodes) == 1:
                pre: ModuleNode = node.prev_nodes[0]
                if node.in_channels != pre.out_channels:
                    assert node.in_channels % pre.out_channels == 0
                    pass_node = self._new_placeholder_node(
                        'pass_placeholder',
                        node.in_channels // pre.out_channels)
                    self._insert_node_before(node, pass_node)

    def _remove_redundant_pass_nodes(self):
        """Remove redundant pass nodes, which do not change number of channels
        and  do not represent any module."""
        for node in copy.copy(list(self.graph.nodes.values())):
            if (node.is_pass_node() and len(node.prev_nodes) == 1
                    and len(node.next_nodes) == 1
                    and not isinstance(node.val, nn.Module)
                    and node.in_channels == node.out_channels):
                self.graph.delete_node(node)

    # topo_rename_nodes
    def _topo_rename(self):
        """Rename cat, bind, pass nodes in topological order."""
        self.cat_placeholder_num = 0
        self.bind_placeholder_num = 0
        self.pass_placeholder_num = 0
        sorted_nodes = OrderedDict()
        for node in self.graph.topo_traverse():
            node: ModuleNode
            if isinstance(node.val, Module):
                pass
            elif node.is_pass_node():
                node.name = f'pass_{self.pass_placeholder_num}'
                self.pass_placeholder_num += 1
            elif node.is_cat_node():
                node.name = f'cat_{self.cat_placeholder_num}'
                self.cat_placeholder_num += 1
            elif node.is_bind_node():
                node.name = f'bind_{self.bind_placeholder_num}'
                self.bind_placeholder_num += 1
            else:
                pass
            sorted_nodes[node.name] = node
        self.graph.nodes = sorted_nodes

    # other
    def _post_process(self):
        """Some post process after init a basic module graph."""
        self._remove_redundant_pass_nodes()
        self._insert_bind_nodes()
        self._insert_pass_nodes()
        self._topo_rename()


class PathToGraphConverter(GraphConverter):
    """The class converts pathlist, which is generated by backward tracer, to a
    module graph."""

    def __init__(self, path_list: PathList, model: Module) -> None:
        """
            Args:
                path_list (PathList): path_list generated by backward tracer.
                model (Module): the model corresponding to the path_list
        """
        super().__init__()
        self.path_list = path_list
        self.cat_dict: Dict[str, str] = {}
        self.name2module = dict(model.named_modules())
        self._pass(self.path_list)

        self._post_process()

    def _pass(self, path_list: PathList):
        """Parse path list."""
        self._parse_helper(path_list, [])

    def _parse_helper(self, path_unit: Union[PathList, Path, PathNode],
                      next_nodes: List[ModuleNode]):
        """Parse a node(unit) in path list."""
        current_node = None
        # path_list
        if isinstance(path_unit, PathList):
            for single_path in path_unit:  # sibling
                self._parse_helper(single_path, next_nodes)

        # path:
        elif isinstance(path_unit, Path):
            current_nexts = next_nodes
            for node in path_unit:  # parent -> children
                current_node = self._parse_helper(node, current_nexts)
                current_nexts = [current_node]

        # Node
        elif isinstance(path_unit, PathNode):

            # cat node: [cat_path_lists]
            if isinstance(path_unit, PathConcatNode):
                current_node = self._add_or_find_node(path_unit)
                self._connect_nexts(current_node, next_nodes)
                for catpath in path_unit.path_lists:  # sibling
                    self._parse_helper(catpath, [current_node])

            # single node
            else:
                current_node = self._add_or_find_node(path_unit)
                self._connect_nexts(current_node, next_nodes)
        return current_node

    def _add_or_find_cat_node(self, pathnode: PathConcatNode):
        """Receive a cat-node.

        If the cat-node exists in the graph, the corresponding node is
        returned, or a new cat node is added to the graph.
        """

        def unify_cat_name(name: str):
            cat_name = name.split('_')
            inputs = sorted(cat_name[1:])
            return f"cat_{'_'.join(inputs)}"

        name_id = pathnode.name
        name_id = unify_cat_name(name_id)
        if name_id in self.cat_dict:
            name = self.cat_dict[name_id]
        else:
            name = f'cat_{self.cat_placeholder_num}'
            self.cat_placeholder_num += 1
            self.cat_dict[name_id] = name
        node = self.graph.add_or_find_node(ModuleNode(name, 'cat_placeholder'))
        return node

    def _add_or_find_node(self, pathnode: PathNode) -> Module:
        """Receive a cat-node.

        If the cat-node exists in the graph, the corresponding node is
        returned, or a new cat node is added to the graph.
        """
        if isinstance(pathnode, PathConcatNode):
            return self._add_or_find_cat_node(pathnode)
        else:
            name = pathnode.name
            assert name in self.name2module, f"{name} doesn't exist in model"
            module = self.name2module[name]
            return self.graph.add_or_find_node(ModuleNode(name, module))

    def _connect_nexts(self, node, nexts: List[ModuleNode]):
        """Connext the node and the nodes in nexts."""
        for next in nexts:
            self.graph.connect(node, next)



# Copyright (c) OpenMMLab. All rights reserved.
def get_placeholder(string: str) -> object:
    """Get placeholder instance which can avoid raising errors when down-stream
    dependency is not installed properly.

    Args:
        string (str): the dependency's name, i.e. `mmcls`
    Raises:
        ImportError: raise it when the dependency is not installed properly.
    Returns:
        object: PlaceHolder instance.
    """

    class PlaceHolder:

        def __init__(self, *args, **kwargs) -> None:
            raise ImportError(
                f'`{string}` is not installed properly, plz check.')

    return PlaceHolder



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional

import torch
from mmengine.structures import BaseDataElement
from torch import nn

from mmrazor.models.utils import add_prefix
from mmrazor.registry import MODELS
from ...base import BaseAlgorithm, LossResults


@MODELS.register_module()
class SelfDistill(BaseAlgorithm):
    """``SelfDistill`` can be used to develop distill algorithms without
    teacher.

    Args:
        distiller (dict): The config dict for built distiller. Distiller may
            have teacher.
        student_trainable (bool): Whether the student is trainable. Defaults
            to True.
        calculate_student_loss (bool): Whether to calculate student loss
            (original task loss) to update student model. Defaults to True.
    """

    def __init__(self,
                 distiller: dict,
                 student_trainable: bool = True,
                 calculate_student_loss: bool = True,
                 **kwargs) -> None:
        super().__init__(**kwargs)

        self.distiller = MODELS.build(distiller)
        # The student model will not calculate gradients and update parameters
        # in some pretraining process.
        self.student_trainable = student_trainable

        # The student loss will not be updated into ``losses`` in some
        # pretraining process.
        self.calculate_student_loss = calculate_student_loss

        # In ``ConfigurableDistller``, the recorder manager is just
        # constructed, but not really initialized yet.
        self.distiller.prepare_from_student(self.student)
        # Still prepare from self-teacher. Teacher recorders of
        # ``SelfDistiller`` hook from self.student but require detach().
        self.distiller.prepare_from_teacher(self.student)

    @property
    def student(self) -> nn.Module:
        """Alias for ``architecture``."""
        return self.architecture

    def loss(
        self,
        batch_inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
    ) -> LossResults:
        """Calculate losses from a batch of inputs and data samples."""

        losses = dict()

        # If the `override_data` of a delivery is True, the delivery will
        # override the origin data with the recorded data.
        self.distiller.set_deliveries_override(True)
        # Original task loss will not be used during some pretraining process.
        if self.calculate_student_loss:
            # teacher_recorders hook from student
            with self.distiller.student_recorders, \
                    self.distiller.teacher_recorders, \
                    self.distiller.deliveries:
                student_losses = self.student(
                    batch_inputs, data_samples, mode='loss')
            losses.update(add_prefix(student_losses, 'student'))
        else:
            with self.distiller.student_recorders, \
                    self.distiller.teacher_recorders, \
                    self.distiller.deliveries:
                if self.student_trainable:
                    _ = self.student(batch_inputs, data_samples, mode='loss')
                else:
                    with torch.no_grad():
                        _ = self.student(
                            batch_inputs, data_samples, mode='loss')

        # Automatically compute distill losses based on `loss_forward_mappings`
        # The required data already exists in the recorders.
        distill_losses = self.distiller.compute_distill_losses()
        losses.update(add_prefix(distill_losses, 'distill'))

        return losses



# Copyright (c) OpenMMLab. All rights reserved.
from .darts_backbone import DartsBackbone
from .searchable_mobilenet import SearchableMobileNet
from .searchable_shufflenet_v2 import SearchableShuffleNetV2
from .wideresnet import WideResNet

__all__ = [
    'SearchableMobileNet', 'SearchableShuffleNetV2', 'DartsBackbone',
    'WideResNet'
]



# Copyright (c) OpenMMLab. All rights reserved.
from .base import DynamicOP
from .default_dynamic_ops import (DynamicBatchNorm, DynamicConv2d,
                                  DynamicGroupNorm, DynamicInstanceNorm,
                                  DynamicLinear)
from .slimmable_dynamic_ops import SwitchableBatchNorm2d

__all__ = [
    'DynamicConv2d', 'DynamicLinear', 'DynamicBatchNorm',
    'DynamicInstanceNorm', 'DynamicGroupNorm', 'SwitchableBatchNorm2d',
    'DynamicOP'
]



# Copyright (c) OpenMMLab. All rights reserved.
from .misc import find_latest_checkpoint
from .placeholder import get_placeholder
from .setup_env import register_all_modules, setup_multi_processes
from .typing import (FixMutable, MultiMutatorsRandomSubnet,
                     SingleMutatorRandomSubnet, SupportRandomSubnet,
                     ValidFixMutable)

__all__ = [
    'find_latest_checkpoint', 'setup_multi_processes', 'register_all_modules',
    'FixMutable', 'ValidFixMutable', 'SingleMutatorRandomSubnet',
    'MultiMutatorsRandomSubnet', 'SupportRandomSubnet', 'get_placeholder'
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Tuple

import torch
import torch.nn as nn

from mmrazor.registry import MODELS


def mask_l2_loss(
        tensor_a: torch.Tensor,
        tensor_b: torch.Tensor,
        saptial_attention_mask: Optional[torch.Tensor] = None,
        channel_attention_mask: Optional[torch.Tensor] = None) -> torch.Tensor:
    """L2 loss with two attention mask, which used to weight the feature
    distillation loss in FBKD.

    Args:
        tensor_a (torch.Tensor): Student featuremap.
        tensor_b (torch.Tensor): Teacher featuremap.
        saptial_attention_mask (torch.Tensor, optional): Mask of spatial-wise
            attention. Defaults to None.
        channel_attention_mask (torch.Tensor, optional): Mask of channel-wise
            attention. Defaults to None.

    Returns:
        diff (torch.Tensor): l2 loss with two attention mask.
    """
    diff = (tensor_a - tensor_b)**2
    if saptial_attention_mask is not None:
        diff = diff * saptial_attention_mask
    if channel_attention_mask is not None:
        diff = diff * channel_attention_mask
    diff = torch.sum(diff)**0.5
    return diff


@MODELS.register_module()
class FBKDLoss(nn.Module):
    """Loss For FBKD, which includs feat_loss, channel_loss, spatial_loss and
    nonlocal_loss.

    Source code:
    https://github.com/ArchipLab-LinfengZhang/Object-Detection-Knowledge-
    Distillation-ICLR2021

    Args:
        mask_l2_weight (float): The weight of the mask l2 loss.
            Defaults to 7e-5, which is the default value in source code.
        channel_weight (float): The weight of the channel loss.
            Defaults to 4e-3, which is the default value in source code.
        spatial_weight (float): The weight of the spatial loss.
            Defaults to 4e-3, which is the default value in source code.
        nonloacl_weight (float): The weight of the nonlocal loss.
            Defaults to 7e-5, which is the default value in source code.
        loss_weight (float): Weight of loss. Defaults to 1.0.
    """

    def __init__(self,
                 mask_l2_weight: float = 7e-5,
                 channel_weight: float = 4e-3,
                 spatial_weight: float = 4e-3,
                 nonloacl_weight: float = 7e-5,
                 loss_weight: float = 1.0) -> None:
        """Inits FBKDLoss."""
        super().__init__()

        self.mask_l2_weight = mask_l2_weight
        self.channel_weight = channel_weight
        self.spatial_weight = spatial_weight
        self.nonloacl_weight = nonloacl_weight
        self.loss_weight = loss_weight

    def forward(self, s_input: Tuple[torch.Tensor, ...],
                t_input: Tuple[torch.Tensor, ...]) -> torch.Tensor:
        """Forward function of FBKDLoss, including feat_loss, channel_loss,
        spatial_loss and nonlocal_loss.

        Args:
            s_input (Tuple[torch.Tensor, ...]): Student input which is the
                output of ``'FBKDStudentConnector'``.
            t_input (Tuple[torch.Tensor, ...]): Teacher input which is the
                output of ``'FBKDTeacherConnector'``.
        """
        losses = 0.0

        (s_spatial_mask, s_channel_mask, s_channel_pool_adapt,
         s_spatial_pool_adapt, s_relation_adapt, s_feat_adapt) = s_input

        (t_spatial_mask, t_channel_mask, t_spatial_pool, t_relation,
         t_feat) = t_input

        # Spatial-wise mask.
        spatial_sum_mask = (t_spatial_mask + s_spatial_mask) / 2
        spatial_sum_mask = spatial_sum_mask.detach()

        # Channel-wise mask, but not used in the FBKD source code.
        channel_sum_mask = (t_channel_mask + s_channel_mask) / 2
        channel_sum_mask = channel_sum_mask.detach()

        # feat_loss with mask
        losses += mask_l2_loss(
            t_feat,
            s_feat_adapt,
            saptial_attention_mask=spatial_sum_mask,
            channel_attention_mask=None) * self.mask_l2_weight

        # channel_loss
        losses += torch.dist(torch.mean(t_feat, [2, 3]),
                             s_channel_pool_adapt) * self.channel_weight

        # spatial_loss
        losses += torch.dist(t_spatial_pool,
                             s_spatial_pool_adapt) * self.spatial_weight

        # nonlocal_loss
        losses += torch.dist(
            t_relation, s_relation_adapt, p=2) * self.nonloacl_weight

        return self.loss_weight * losses



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Dict

import torch.nn as nn

from mmrazor.models.mutables.mutable_channel import MutableChannel
from mmrazor.registry import MODELS
from .base import DynamicOP


class SwitchableBatchNorm2d(nn.Module, DynamicOP):
    """Employs independent batch normalization for different switches in a
    slimmable network.

    To train slimmable networks, ``SwitchableBatchNorm2d`` privatizes all
    batch normalization layers for each switch in a slimmable network.
    Compared with the naive training approach, it solves the problem of feature
    aggregation inconsistency between different switches by independently
    normalizing the feature mean and variance during testing.

    Args:
        module_name (str): Name of this `SwitchableBatchNorm2d`.
        num_features_cfg (Dict): Config related to `num_features`.
        eps (float): A value added to the denominator for numerical stability.
            Same as that in :obj:`torch.nn._BatchNorm`. Default: 1e-5
        momentum (float): The value used for the running_mean and running_var
            computation. Can be set to None for cumulative moving average
            (i.e. simple average). Same as that in :obj:`torch.nn._BatchNorm`.
            Default: 0.1
        affine (bool): A boolean value that when set to True, this module has
            learnable affine parameters. Same as that in
            :obj:`torch.nn._BatchNorm`. Default: True
        track_running_stats (bool): A boolean value that when set to True, this
            module tracks the running mean and variance, and when set to False,
            this module does not track such statistics, and initializes
            statistics buffers running_mean and running_var as None. When these
            buffers are None, this module always uses batch statistics. in both
            training and eval modes. Same as that in
            :obj:`torch.nn._BatchNorm`. Default: True
    """

    def __init__(self,
                 num_features_cfg: Dict,
                 eps: float = 1e-5,
                 momentum: float = 0.1,
                 affine: bool = True,
                 track_running_stats: bool = True):
        super(SwitchableBatchNorm2d, self).__init__()

        num_features_cfg = copy.deepcopy(num_features_cfg)
        candidate_choices = num_features_cfg.pop('candidate_choices')
        num_features_cfg.update(dict(num_channels=max(candidate_choices)))

        bns = [
            nn.BatchNorm2d(num_features, eps, momentum, affine,
                           track_running_stats)
            for num_features in candidate_choices
        ]
        self.bns = nn.ModuleList(bns)

        self.mutable_num_features = MODELS.build(num_features_cfg)

    @property
    def mutable_in(self) -> MutableChannel:
        """Mutable `num_features`."""
        return self.mutable_num_features

    @property
    def mutable_out(self) -> MutableChannel:
        """Mutable `num_features`."""
        return self.mutable_num_features

    def forward(self, input):
        """Forward computation according to the current switch of the slimmable
        networks."""
        idx = self.mutable_num_features.current_choice
        return self.bns[idx](input)

    def to_static_op(self) -> nn.Module:
        bn_idx = self.mutable_num_features.current_choice

        return self.bns[bn_idx]



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from pathlib import Path
from typing import Optional, Sequence, Union

from mmengine.dist import master_only
from mmengine.fileio import FileClient, dump
from mmengine.hooks import Hook
from mmengine.registry import HOOKS

DATA_BATCH = Optional[Sequence[dict]]


@HOOKS.register_module()
class DumpSubnetHook(Hook):
    """Dump subnet periodically.

    Args:
        interval (int): The saving period. If ``by_epoch=True``, interval
            indicates epochs, otherwise it indicates iterations.
            Defaults to -1, which means "never".
        by_epoch (bool): Saving checkpoints by epoch or by iteration.
            Default: True.
        out_dir (str, optional | Path): The root directory to save checkpoints.
            If not specified, ``runner.work_dir`` will be used by default. If
            specified, the ``out_dir`` will be the concatenation of ``out_dir``
            and the last level directory of ``runner.work_dir``. For example,
            if the input ``our_dir`` is ``./tmp`` and ``runner.work_dir`` is
            ``./work_dir/cur_exp``, then the ckpt will be saved in
            ``./tmp/cur_exp``. Defaults to None.
        max_keep_subnets (int): The maximum subnets to keep.
            In some cases we want only the latest few subnets and would
            like to delete old ones to save the disk space.
            Defaults to -1, which means unlimited.
        save_last (bool): Whether to force the last checkpoint to be
            saved regardless of interval. Defaults to True.
        file_client_args (dict, optional): Arguments to instantiate a
            FileClient. See :class:`mmcv.fileio.FileClient` for details.
            Defaults to None.
    """
    out_dir: str

    priority = 'VERY_LOW'

    def __init__(self,
                 interval: int = -1,
                 by_epoch: bool = True,
                 out_dir: Optional[Union[str, Path]] = None,
                 max_keep_subnets: int = -1,
                 save_last: bool = True,
                 file_client_args: Optional[dict] = None,
                 **kwargs) -> None:
        self.interval = interval
        self.by_epoch = by_epoch
        self.out_dir = out_dir  # type: ignore
        self.max_keep_subnets = max_keep_subnets
        self.save_last = save_last
        self.args = kwargs
        self.file_client_args = file_client_args

    def before_train(self, runner) -> None:
        """Finish all operations, related to checkpoint.

        This function will get the appropriate file client, and the directory
        to save these checkpoints of the model.

        Args:
            runner (Runner): The runner of the training process.
        """
        if self.out_dir is None:
            self.out_dir = runner.work_dir

        self.file_client = FileClient.infer_client(self.file_client_args,
                                                   self.out_dir)
        # if `self.out_dir` is not equal to `runner.work_dir`, it means that
        # `self.out_dir` is set so the final `self.out_dir` is the
        # concatenation of `self.out_dir` and the last level directory of
        # `runner.work_dir`
        if self.out_dir != runner.work_dir:
            basename = osp.basename(runner.work_dir.rstrip(osp.sep))
            self.out_dir = self.file_client.join_path(
                self.out_dir, basename)  # type: ignore  # noqa: E501

        runner.logger.info(f'Subnets will be saved to {self.out_dir} by '
                           f'{self.file_client.name}.')

    def after_train_epoch(self, runner) -> None:
        """Save the checkpoint and synchronize buffers after each epoch.

        Args:
            runner (Runner): The runner of the training process.
        """
        if not self.by_epoch:
            return

        # save subnet for following cases:
        # 1. every ``self.interval`` epochs
        # 2. reach the last epoch of training
        if self.every_n_epochs(runner, self.interval) or (
                self.save_last and self.is_last_train_epoch(runner)):
            runner.logger.info(f'Saving subnet at {runner.epoch + 1} epochs')
            self._save_subnet(runner)

    @master_only
    def _save_subnet(self, runner) -> None:
        """Save the current subnet and delete outdated subnet.

        Args:
            runner (Runner): The runner of the training process.
        """

        if runner.distributed:
            subnet_dict = runner.model.module.search_subnet()
        else:
            subnet_dict = runner.model.search_subnet()

        if self.by_epoch:
            subnet_filename = self.args.get(
                'filename_tmpl',
                'subnet_epoch_{}.yaml').format(runner.epoch + 1)
        else:
            subnet_filename = self.args.get(
                'filename_tmpl', 'subnet_iter_{}.yaml').format(runner.iter + 1)

        file_client = FileClient.infer_client(self.file_client_args,
                                              self.out_dir)
        filepath = file_client.join_path(self.out_dir, subnet_filename)

        dump(subnet_dict, filepath, file_format='yaml')

    def after_train_iter(self,
                         runner,
                         batch_idx: int,
                         data_batch: DATA_BATCH = None,
                         outputs=Optional[dict]) -> None:
        """Save the subnet after each iteration.

        Args:
            runner (Runner): The runner of the training process.
            batch_idx (int): The index of the current batch in the train loop.
            data_batch (Sequence[dict], optional): Data from dataloader.
                Defaults to None.
            outputs (dict, optional): Outputs from model. Defaults to None.
        """
        if self.by_epoch:
            return

        # save checkpoint for following cases:
        # 1. every ``self.interval`` iterations
        # 2. reach the last iteration of training
        if self.every_n_train_iters(runner, self.interval) or \
                (self.save_last and
                 self.is_last_train_iter(runner)):
            runner.logger.info(
                f'Saving subnet at {runner.iter + 1} iterations')
            self._save_subnet(runner)



# Copyright (c) OpenMMLab. All rights reserved.
import torch
from mmdet.models import BaseDetector

from mmrazor.registry import TASK_UTILS


# todo: adapt to mmdet 2.0
@TASK_UTILS.register_module()
class SingleStageDetectorPseudoLoss:

    def __call__(self, model: BaseDetector) -> torch.Tensor:
        pseudo_img = torch.rand(1, 3, 224, 224)
        pseudo_output = model.forward_dummy(pseudo_img)
        out = torch.tensor(0.)
        for levels in pseudo_output:
            out += sum([level.sum() for level in levels])

        return out



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Callable, Dict, List, Optional, Union

import numpy as np
import torch

from mmrazor.registry import MODELS
from ..derived_mutable import DerivedMutable
from .mutable_channel import MutableChannel


@MODELS.register_module()
class OneShotMutableChannel(MutableChannel[int, Dict]):
    """A type of ``MUTABLES`` for single path supernet such as AutoSlim. In
    single path supernet, each module only has one choice invoked at the same
    time. A path is obtained by sampling all the available choices. It is the
    base class for one shot mutable channel.

    Args:
        num_channels (int): The raw number of channels.
        candidate_choices (List): If `candidate_mode` is "ratio",
            candidate_choices is a list of candidate width ratios. If
            `candidate_mode` is "number", candidate_choices is a list of
            candidate channel number. We note that the width ratio is the ratio
            between the number of reserved channels and that of all channels in
            a layer.
            For example, if `ratios` is [0.25, 0.5], there are 2 cases
            for us to choose from when we sample from a layer with 12 channels.
            One is sampling the very first 3 channels in this layer, another is
            sampling the very first 6 channels in this layer.
        candidate_mode (str): One of "ratio" or "number".
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.
    """

    def __init__(self,
                 num_channels: int,
                 candidate_choices: List[Union[int, float]],
                 candidate_mode: str = 'ratio',
                 init_cfg: Optional[Dict] = None):
        super(OneShotMutableChannel, self).__init__(
            num_channels=num_channels, init_cfg=init_cfg)

        self._current_choice = num_channels
        assert len(candidate_choices) > 0, \
            f'Number of candidate choices must be greater than 0, ' \
            f'but got: {len(candidate_choices)}'
        self._candidate_choices = candidate_choices
        assert candidate_mode in ['ratio', 'number']
        self._candidate_mode = candidate_mode

        self._check_candidate_choices()

    def _check_candidate_choices(self):
        """Check if the input `candidate_choices` is valid."""
        if self._candidate_mode == 'number':
            assert all([num > 0 and num <= self.num_channels
                        for num in self._candidate_choices]), \
                f'The candidate channel numbers should be in ' \
                f'range(0, {self.num_channels}].'
            assert all([isinstance(num, int)
                        for num in self._candidate_choices]), \
                'Type of `candidate_choices` should be int.'
        else:
            assert all([
                ratio > 0 and ratio <= 1 for ratio in self._candidate_choices
            ]), 'The candidate ratio should be in range(0, 1].'

    def sample_choice(self) -> int:
        """Sample an arbitrary selection from candidate choices.

        Returns:
            int: The chosen number of channels.
        """
        assert len(self.concat_parent_mutables) == 0
        num_channels = np.random.choice(self.choices)
        assert num_channels > 0, \
            f'Sampled number of channels in `Mutable` {self.name}' \
            f' should be a positive integer.'
        return num_channels

    @property
    def min_choice(self) -> int:
        """Minimum number of channels."""
        assert len(self.concat_parent_mutables) == 0
        min_channels = min(self.choices)
        assert min_channels > 0, \
            f'Minimum number of channels in `Mutable` {self.name}' \
            f' should be a positive integer.'
        return min_channels

    @property
    def max_choice(self) -> int:
        """Maximum number of channels."""
        return max(self.choices)

    @property
    def current_choice(self):
        """The current choice of the mutable."""
        assert len(self.concat_parent_mutables) == 0
        return self._current_choice

    @current_choice.setter
    def current_choice(self, choice: int):
        """Set the current choice of the mutable."""
        assert choice in self.choices
        self._current_choice = choice

    @property
    def choices(self) -> List:
        """list: all choices. """
        if self._candidate_mode == 'number':
            return self._candidate_choices
        candidate_choices = [
            round(ratio * self.num_channels)
            for ratio in self._candidate_choices
        ]
        return candidate_choices

    @property
    def num_choices(self) -> int:
        return len(self.choices)

    def convert_choice_to_mask(self, choice: int) -> torch.Tensor:
        """Get the mask according to the input choice."""
        num_channels = choice
        mask = torch.zeros(self.num_channels).bool()
        mask[:num_channels] = True
        return mask

    def dump_chosen(self) -> Dict:
        assert self.current_choice is not None

        return dict(
            current_choice=self.current_choice,
            origin_channels=self.num_channels)

    def fix_chosen(self, dumped_chosen: Dict) -> None:
        if self.is_fixed:
            raise RuntimeError('OneShotMutableChannel can not be fixed twice')

        current_choice = dumped_chosen['current_choice']
        origin_channels = dumped_chosen['origin_channels']

        assert current_choice <= origin_channels
        assert origin_channels == self.num_channels

        self.current_choice = current_choice
        self.is_fixed = True

    def __repr__(self):
        concat_mutable_name = [
            mutable.name for mutable in self.concat_parent_mutables
        ]
        repr_str = self.__class__.__name__
        repr_str += f'(name={self.name}, '
        repr_str += f'num_channels={self.num_channels}, '
        repr_str += f'current_choice={self.current_choice}, '
        repr_str += f'choices={self.choices}, '
        repr_str += f'activated_channels={self.current_mask.sum().item()}, '
        repr_str += f'concat_mutable_name={concat_mutable_name})'
        return repr_str

    def __rmul__(self, other) -> DerivedMutable:
        return self * other

    def __mul__(self, other) -> DerivedMutable:
        if isinstance(other, int):
            return self.derive_expand_mutable(other)

        from ..mutable_value import OneShotMutableValue

        def expand_choice_fn(mutable1: 'OneShotMutableChannel',
                             mutable2: OneShotMutableValue) -> Callable:

            def fn():
                return mutable1.current_choice * mutable2.current_choice

            return fn

        def expand_mask_fn(mutable1: 'OneShotMutableChannel',
                           mutable2: OneShotMutableValue) -> Callable:

            def fn():
                mask = mutable1.current_mask
                max_expand_ratio = mutable2.max_choice
                current_expand_ratio = mutable2.current_choice
                expand_num_channels = mask.size(0) * max_expand_ratio

                expand_choice = mutable1.current_choice * current_expand_ratio
                expand_mask = torch.zeros(expand_num_channels).bool()
                expand_mask[:expand_choice] = True

                return expand_mask

            return fn

        if isinstance(other, OneShotMutableValue):
            return DerivedMutable(
                choice_fn=expand_choice_fn(self, other),
                mask_fn=expand_mask_fn(self, other))

        raise TypeError(f'Unsupported type {type(other)} for mul!')

    def __floordiv__(self, other) -> DerivedMutable:
        if isinstance(other, int):
            return self.derive_divide_mutable(other)
        if isinstance(other, tuple):
            assert len(other) == 2
            return self.derive_divide_mutable(*other)

        raise TypeError(f'Unsupported type {type(other)} for div!')



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F

from mmrazor.registry import MODELS


@MODELS.register_module()
class ATLoss(nn.Module):
    """"Paying More Attention to Attention: Improving the Performance of
    Convolutional Neural Networks via Attention Transfer" Conference paper at
    ICLR2017 https://openreview.net/forum?id=Sks9_ajex.

    https://github.com/szagoruyko/attention-transfer/blob/master/utils.py

    Args:
        loss_weight (float): Weight of loss. Defaults to 1.0.
    """

    def __init__(
        self,
        loss_weight: float = 1.0,
    ) -> None:
        super().__init__()
        self.loss_weight = loss_weight

    def forward(self, s_feature: torch.Tensor,
                t_feature: torch.Tensor) -> torch.Tensor:
        """"Forward function for ATLoss."""
        loss = (self.calc_attention_matrix(s_feature) -
                self.calc_attention_matrix(t_feature)).pow(2).mean()
        return self.loss_weight * loss

    def calc_attention_matrix(self, x: torch.Tensor) -> torch.Tensor:
        """"Calculate the attention matrix.

        Args:
            x (torch.Tensor): Input features.
        """
        return F.normalize(x.pow(2).mean(1).view(x.size(0), -1))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional

import torch

from mmrazor.registry import MODELS
from .mutable_channel import MutableChannel


@MODELS.register_module()
class SlimmableMutableChannel(MutableChannel[int, Dict[str, int]]):
    """A type of ``MUTABLES`` to train several subnet together, such as the
    retraining stage in AutoSlim.

    Notes:
        We need to set `candidate_choices` after the instantiation of a
        `SlimmableMutableChannel` by ourselves.

    Args:
        num_channels (int): The raw number of channels.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.
    """

    def __init__(self, num_channels: int, init_cfg: Optional[Dict] = None):
        super(SlimmableMutableChannel, self).__init__(
            num_channels=num_channels, init_cfg=init_cfg)

        self.num_channels = num_channels

    @property
    def candidate_choices(self) -> List:
        """A list of candidate channel numbers."""
        return self._candidate_choices

    @candidate_choices.setter
    def candidate_choices(self, choices):
        """Set the candidate channel numbers."""
        assert getattr(self, '_candidate_choices', None) is None, \
            f'candidate_choices can be set only when candidate_choices is ' \
            f'None, got: candidate_choices = {self._candidate_choices}'

        assert all([num > 0 and num <= self.num_channels
                    for num in choices]), \
            f'The candidate channel numbers should be in ' \
            f'range(0, {self.num_channels}].'
        assert all([isinstance(num, int) for num in choices]), \
            'Type of `candidate_choices` should be int.'

        self._candidate_choices = list(choices)

    @property
    def choices(self) -> List[int]:
        """Return all subnet indexes."""
        assert self._candidate_choices is not None
        return list(range(len(self.candidate_choices)))

    def dump_chosen(self) -> Dict:
        assert self.current_choice is not None

        return dict(
            current_choice=self._candidate_choices[self.current_choice],
            origin_channels=self.num_channels)

    def fix_chosen(self, dumped_chosen: Dict) -> None:
        chosen = dumped_chosen['current_choice']
        origin_channels = dumped_chosen['origin_channels']

        assert chosen <= origin_channels

        # TODO
        # remove after remove `current_choice`
        self.current_choice = self.candidate_choices.index(chosen)
        self._chosen = chosen

        super().fix_chosen(chosen)

    @property
    def num_choices(self) -> int:
        return len(self.choices)

    def convert_choice_to_mask(self, choice: int) -> torch.Tensor:
        """Get the mask according to the input choice."""
        if self.is_fixed:
            num_channels = self._chosen
        elif not hasattr(self, '_candidate_choices'):
            # todo: we trace the supernet before set_candidate_choices.
            #  It's hacky
            num_channels = self.num_channels
        else:
            num_channels = self.candidate_choices[choice]
        mask = torch.zeros(self.num_channels).bool()
        mask[:num_channels] = True
        return mask



# dataset settings
dataset_type = 'mmcls.CIFAR10'
preprocess_cfg = dict(
    # RGB format normalization parameters
    mean=[125.307, 122.961, 113.8575],
    std=[51.5865, 50.847, 51.255],
    # loaded images are already RGB format
    to_rgb=False)

train_pipeline = [
    dict(type='mmcls.RandomCrop', crop_size=32, padding=4),
    dict(type='mmcls.RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='mmcls.Cutout', shape=16, pad_val=0, prob=1.0),
    dict(type='mmcls.PackClsInputs'),
]

test_pipeline = [
    dict(type='mmcls.PackClsInputs'),
]

train_dataloader = dict(
    batch_size=96,
    num_workers=2,
    dataset=dict(
        type=dataset_type,
        data_prefix='data/cifar10',
        test_mode=False,
        pipeline=train_pipeline),
    sampler=dict(type='mmcls.DefaultSampler', shuffle=True),
    persistent_workers=True,
)

val_dataloader = dict(
    batch_size=16,
    num_workers=2,
    dataset=dict(
        type=dataset_type,
        data_prefix='data/cifar10/',
        test_mode=True,
        pipeline=test_pipeline),
    sampler=dict(type='DefaultSampler', shuffle=False),
    persistent_workers=True,
)
val_evaluator = dict(type='mmcls.Accuracy', topk=(1, ))

test_dataloader = val_dataloader
test_evaluator = val_evaluator

# optimizer
optim_wrapper = dict(
    optimizer=dict(type='SGD', lr=0.025, momentum=0.9, weight_decay=3e-4),
    clip_grad=dict(max_norm=5, norm_type=2))

# leanring policy
param_scheduler = [
    dict(
        type='CosineAnnealingLR',
        T_max=600,
        by_epoch=True,
        begin=0,
        end=600,
    )
]

# train, val, test setting
train_cfg = dict(by_epoch=True, max_epochs=600)
val_cfg = dict()  # validate each epoch
test_cfg = dict()  # dataset settings



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Union

from mmengine.evaluator import Evaluator
from mmengine.runner import ValLoop
from torch.utils.data import DataLoader

from mmrazor.models.utils import add_prefix
from mmrazor.registry import LOOPS


@LOOPS.register_module()
class SlimmableValLoop(ValLoop):
    """Knowledge Distill loop for validation. It is not only validate student,
    but also validate teacher with the same dataloader.

    Args:
        runner (Runner): A reference of runner.
        dataloader (Dataloader or dict): A dataloader object or a dict to
            build a dataloader.
        evaluator (Evaluator or dict or list): Used for computing metrics.
        fp16 (bool): Whether to enable fp16 validation. Defaults to
            False.
    """

    def __init__(self,
                 runner,
                 dataloader: Union[DataLoader, Dict],
                 evaluator: Union[Evaluator, Dict, List],
                 fp16: bool = False) -> None:
        super().__init__(runner, dataloader, evaluator, fp16)

        if self.runner.distributed:
            model = self.runner.model.module
        else:
            model = self.runner.model

        # just for convenience
        self._model = model

    def run(self):
        """Launch validation."""
        self.runner.call_hook('before_val')

        all_metrics = dict()
        for subnet_idx in range(self._model.num_subnet):
            self.runner.call_hook('before_val_epoch')
            self.runner.model.eval()
            self._model.mutator.switch_choices(subnet_idx)
            for idx, data_batch in enumerate(self.dataloader):
                self.run_iter(idx, data_batch)
            # compute student metrics
            metrics = self.evaluator.evaluate(len(self.dataloader.dataset))
            all_metrics.update(add_prefix(metrics, f'subnet_{subnet_idx}'))

        self.runner.call_hook('after_val_epoch', metrics=all_metrics)

        self.runner.call_hook('after_val')



_base_ = [
    'mmdet::_base_/datasets/coco_detection.py',
    'mmdet::_base_/schedules/schedule_1x.py',
    'mmdet::_base_/default_runtime.py'
]

teacher_ckpt = 'https://download.openmmlab.com/mmdetection/v2.0/faster_rcnn/faster_rcnn_r101_fpn_1x_coco/faster_rcnn_r101_fpn_1x_coco_20200130-f513f705.pth'  # noqa: E501
model = dict(
    _scope_='mmrazor',
    type='SingleTeacherDistill',
    architecture=dict(
        cfg_path='mmdet::faster_rcnn/faster-rcnn_r50_fpn_1x_coco.py',
        pretrained=True),
    teacher=dict(
        cfg_path='mmdet::faster_rcnn/faster-rcnn_r101_fpn_1x_coco.py',
        pretrained=False),
    teacher_ckpt=teacher_ckpt,
    distiller=dict(
        type='ConfigurableDistiller',
        student_recorders=dict(
            neck_s0=dict(type='ModuleOutputs', source='neck.fpn_convs.0.conv'),
            neck_s1=dict(type='ModuleOutputs', source='neck.fpn_convs.1.conv'),
            neck_s2=dict(type='ModuleOutputs', source='neck.fpn_convs.2.conv'),
            neck_s3=dict(type='ModuleOutputs',
                         source='neck.fpn_convs.3.conv')),
        teacher_recorders=dict(
            neck_s0=dict(type='ModuleOutputs', source='neck.fpn_convs.0.conv'),
            neck_s1=dict(type='ModuleOutputs', source='neck.fpn_convs.1.conv'),
            neck_s2=dict(type='ModuleOutputs', source='neck.fpn_convs.2.conv'),
            neck_s3=dict(type='ModuleOutputs',
                         source='neck.fpn_convs.3.conv')),
        distill_losses=dict(
            loss_s0=dict(type='FBKDLoss'),
            loss_s1=dict(type='FBKDLoss'),
            loss_s2=dict(type='FBKDLoss'),
            loss_s3=dict(type='FBKDLoss')),
        connectors=dict(
            loss_s0_sfeat=dict(
                type='FBKDStudentConnector',
                in_channels=256,
                reduction=4,
                mode='dot_product',
                sub_sample=True,
                maxpool_stride=8),
            loss_s0_tfeat=dict(
                type='FBKDTeacherConnector',
                in_channels=256,
                reduction=4,
                mode='dot_product',
                sub_sample=True,
                maxpool_stride=8),
            loss_s1_sfeat=dict(
                type='FBKDStudentConnector',
                in_channels=256,
                reduction=4,
                mode='dot_product',
                sub_sample=True,
                maxpool_stride=4),
            loss_s1_tfeat=dict(
                type='FBKDTeacherConnector',
                in_channels=256,
                reduction=4,
                mode='dot_product',
                sub_sample=True,
                maxpool_stride=4),
            loss_s2_sfeat=dict(
                type='FBKDStudentConnector',
                in_channels=256,
                mode='dot_product',
                sub_sample=True),
            loss_s2_tfeat=dict(
                type='FBKDTeacherConnector',
                in_channels=256,
                mode='dot_product',
                sub_sample=True),
            loss_s3_sfeat=dict(
                type='FBKDStudentConnector',
                in_channels=256,
                mode='dot_product',
                sub_sample=True),
            loss_s3_tfeat=dict(
                type='FBKDTeacherConnector',
                in_channels=256,
                mode='dot_product',
                sub_sample=True)),
        loss_forward_mappings=dict(
            loss_s0=dict(
                s_input=dict(
                    from_student=True,
                    recorder='neck_s0',
                    connector='loss_s0_sfeat'),
                t_input=dict(
                    from_student=False,
                    recorder='neck_s0',
                    connector='loss_s0_tfeat')),
            loss_s1=dict(
                s_input=dict(
                    from_student=True,
                    recorder='neck_s1',
                    connector='loss_s1_sfeat'),
                t_input=dict(
                    from_student=False,
                    recorder='neck_s1',
                    connector='loss_s1_tfeat')),
            loss_s2=dict(
                s_input=dict(
                    from_student=True,
                    recorder='neck_s2',
                    connector='loss_s2_sfeat'),
                t_input=dict(
                    from_student=False,
                    recorder='neck_s2',
                    connector='loss_s2_tfeat')),
            loss_s3=dict(
                s_input=dict(
                    from_student=True,
                    recorder='neck_s3',
                    connector='loss_s3_sfeat'),
                t_input=dict(
                    from_student=False,
                    recorder='neck_s3',
                    connector='loss_s3_tfeat')))))

find_unused_parameters = True

val_cfg = dict(_delete_=True, type='mmrazor.SingleTeacherDistillValLoop')



# Copyright (c) OpenMMLab. All rights reserved.
from .algorithms import *  # noqa: F401,F403
from .architectures import *  # noqa: F401,F403
from .distillers import *  # noqa: F401,F403
from .losses import *  # noqa: F401,F403
from .mutables import *  # noqa: F401,F403
from .mutators import *  # noqa: F401,F403
from .task_modules import *  # noqa: F401,F403



# Copyright (c) OpenMMLab. All rights reserved.
import sys

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

import inspect
import logging
from itertools import product
from typing import Any, Callable, Dict, Iterable, Optional, Set, Union

import torch
from mmengine.logging import print_log
from torch import Tensor

from ..utils import make_divisible
from .base_mutable import CHOICE_TYPE, BaseMutable


class MutableProtocol(Protocol):  # pragma: no cover
    """Protocol for Mutable."""

    @property
    def current_choice(self) -> Any:
        """Current choice."""

    def derive_expand_mutable(self, expand_ratio: int) -> Any:
        """Derive expand mutable."""

    def derive_divide_mutable(self, ratio: int, divisor: int) -> Any:
        """Derive divide mutable."""


class MutableChannelProtocol(MutableProtocol):  # pragma: no cover
    """Protocol for MutableChannel."""

    @property
    def current_mask(self) -> Tensor:
        """Current mask."""


def _expand_choice_fn(mutable: MutableProtocol, expand_ratio: int) -> Callable:
    """Helper function to build `choice_fn` for expand derived mutable."""

    def fn():
        return mutable.current_choice * expand_ratio

    return fn


def _expand_mask_fn(mutable: MutableProtocol,
                    expand_ratio: int) -> Callable:  # pragma: no cover
    """Helper function to build `mask_fn` for expand derived mutable."""
    if not hasattr(mutable, 'current_mask'):
        raise ValueError('mutable must have attribute `currnet_mask`')

    def fn():
        mask = mutable.current_mask
        expand_num_channels = mask.size(0) * expand_ratio
        expand_choice = mutable.current_choice * expand_ratio
        expand_mask = torch.zeros(expand_num_channels).bool()
        expand_mask[:expand_choice] = True

        return expand_mask

    return fn


def _divide_and_divise(x: int, ratio: int, divisor: int = 8) -> int:
    """Helper function for divide and divise."""
    new_x = x // ratio

    return make_divisible(new_x, divisor)


def _divide_choice_fn(mutable: MutableProtocol,
                      ratio: int,
                      divisor: int = 8) -> Callable:
    """Helper function to build `choice_fn` for divide derived mutable."""

    def fn():
        return _divide_and_divise(mutable.current_choice, ratio, divisor)

    return fn


def _divide_mask_fn(mutable: MutableProtocol,
                    ratio: int,
                    divisor: int = 8) -> Callable:  # pragma: no cover
    """Helper function to build `mask_fn` for divide derived mutable."""
    if not hasattr(mutable, 'current_mask'):
        raise ValueError('mutable must have attribute `currnet_mask`')

    def fn():
        mask = mutable.current_mask
        divide_num_channels = _divide_and_divise(mask.size(0), ratio, divisor)
        divide_choice = _divide_and_divise(mutable.current_choice, ratio,
                                           divisor)
        divide_mask = torch.zeros(divide_num_channels).bool()
        divide_mask[:divide_choice] = True

        return divide_mask

    return fn


def _concat_choice_fn(mutables: Iterable[MutableChannelProtocol]) -> Callable:
    """Helper function to build `choice_fn` for concat derived mutable."""

    def fn():
        return sum((m.current_choice for m in mutables))

    return fn


def _concat_mask_fn(mutables: Iterable[MutableChannelProtocol]) -> Callable:
    """Helper function to build `mask_fn` for concat derived mutable."""

    def fn():
        return torch.cat([m.current_mask for m in mutables])

    return fn


class DerivedMethodMixin:
    """A mixin that provides some useful method to derive mutable."""

    def derive_same_mutable(self: MutableProtocol) -> 'DerivedMutable':
        """Derive same mutable as the source."""
        return self.derive_expand_mutable(expand_ratio=1)

    def derive_expand_mutable(self: MutableProtocol,
                              expand_ratio: int) -> 'DerivedMutable':
        """Derive expand mutable, usually used with `expand_ratio`."""
        choice_fn = _expand_choice_fn(self, expand_ratio=expand_ratio)

        mask_fn: Optional[Callable] = None
        if hasattr(self, 'current_mask'):
            mask_fn = _expand_mask_fn(self, expand_ratio=expand_ratio)

        return DerivedMutable(choice_fn=choice_fn, mask_fn=mask_fn)

    def derive_divide_mutable(self: MutableProtocol,
                              ratio: int,
                              divisor: int = 8) -> 'DerivedMutable':
        """Derive divide mutable, usually used with `make_divisable`."""
        choice_fn = _divide_choice_fn(self, ratio=ratio, divisor=divisor)

        mask_fn: Optional[Callable] = None
        if hasattr(self, 'current_mask'):
            mask_fn = _divide_mask_fn(self, ratio=ratio, divisor=divisor)

        return DerivedMutable(choice_fn=choice_fn, mask_fn=mask_fn)

    @staticmethod
    def derive_concat_mutable(
            mutables: Iterable[MutableChannelProtocol]) -> 'DerivedMutable':
        """Derive concat mutable, usually used with `torch.cat`."""
        for mutable in mutables:
            if not hasattr(mutable, 'current_mask'):
                raise RuntimeError('Source mutable of concat derived mutable '
                                   'must have attribute `currnet_mask`')

        choice_fn = _concat_choice_fn(mutables)
        mask_fn = _concat_mask_fn(mutables)

        return DerivedMutable(choice_fn=choice_fn, mask_fn=mask_fn)


class DerivedMutable(BaseMutable[CHOICE_TYPE, CHOICE_TYPE],
                     DerivedMethodMixin):
    """Class for derived mutable.

    A derived mutable is a mutable derived from other mutables that has
    `current_choice` and `current_mask` attributes (if any).

    Note:
        A derived mutable does not have its own search space, so it is
        not legal to modify its `current_choice` or `current_mask` directly.
        And the only way to modify them is by modifying `current_choice` or
        `current_mask` in corresponding source mutables.

    Args:
        choice_fn (callable): A closure that controls how to generate
            `current_choice`.
        mask_fn (callable, optional): A closure that controls how to generate
            `current_mask`. Defaults to None.
        source_mutables (iterable, optional): Specify source mutables for this
            derived mutable. If the argument is None, source mutables will be
            traced automatically by parsing mutables in closure variables.
            Defaults to None.
        alias (str, optional): alias of the `MUTABLE`. Defaults to None.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`. Defaults to None.

    Examples:
        >>> from mmrazor.models.mutables import OneShotMutableChannel
        >>> mutable_channel = OneShotMutableChannel(
        ...     num_channels=3,
        ...     candidate_choices=[1, 2, 3],
        ...     candidate_mode='number')
        >>> # derive expand mutable
        >>> derived_mutable_channel = mutable_channel * 2
        >>> # source mutables will be traced automatically
        >>> derived_mutable_channel.source_mutables
        {OneShotMutableChannel(name=unbind, num_channels=3, current_choice=3, choices=[1, 2, 3], activated_channels=3, concat_mutable_name=[])}  # noqa: E501
        >>> # modify `current_choice` of `mutable_channel`
        >>> mutable_channel.current_choice = 2
        >>> # `current_choice` and `current_mask` of derived mutable will be modified automatically  # noqa: E501
        >>> derived_mutable_channel
        DerivedMutable(current_choice=4, activated_channels=4, source_mutables={OneShotMutableChannel(name=unbind, num_channels=3, current_choice=2, choices=[1, 2, 3], activated_channels=2, concat_mutable_name=[])}, is_fixed=False)  # noqa: E501
    """

    def __init__(self,
                 choice_fn: Callable,
                 mask_fn: Optional[Callable] = None,
                 source_mutables: Optional[Iterable[BaseMutable]] = None,
                 alias: Optional[str] = None,
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(alias, init_cfg)

        self.choice_fn = choice_fn
        self.mask_fn = mask_fn

        if source_mutables is None:
            source_mutables = self._trace_source_mutables()
            if len(source_mutables) == 0:
                raise RuntimeError(
                    'Can not find source mutables automatically, '
                    'please provide manually.')
        else:
            source_mutables = set(source_mutables)
        for mutable in source_mutables:
            if not self.is_source_mutable(mutable):
                raise ValueError('Expect all mutable to be source mutable, '
                                 f'but {mutable} is not')
        self.source_mutables = source_mutables

    # TODO
    # has no effect
    def fix_chosen(self, chosen: CHOICE_TYPE) -> None:
        """Fix mutable with subnet config.

        Warning:
            Fix derived mutable will have no actually effect.
        """
        print_log(
            'Trying to fix chosen for derived mutable, '
            'which will have no effect.',
            level=logging.WARNING)

    def dump_chosen(self) -> CHOICE_TYPE:
        """Dump information of chosen.

        Returns:
            Dict: Dumped information.
        """
        print_log(
            'Trying to dump chosen for derived mutable, '
            'but its value depend on the source mutables.',
            level=logging.WARNING)
        return self.current_choice

    @property
    def is_fixed(self) -> bool:
        """Whether the derived mutable is fixed.

        Note:
            Depends on whether all source mutables are already fixed.
        """
        return all(m.is_fixed for m in self.source_mutables)

    @is_fixed.setter
    def is_fixed(self, is_fixed: bool) -> bool:
        """Setter of is fixed."""
        raise RuntimeError(
            '`is_fixed` of derived mutable should not be modified directly')

    @property
    def choices(self):
        origin_choices = [m.current_choice for m in self.source_mutables]

        all_choices = [m.choices for m in self.source_mutables]

        product_choices = product(*all_choices)

        derived_choices = list()
        for item_choices in product_choices:
            for m, choice in zip(self.source_mutables, item_choices):
                m.current_choice = choice

            derived_choices.append(self.choice_fn())

        for m, choice in zip(self.source_mutables, origin_choices):
            m.current_choice = choice

        return derived_choices

    @property
    def num_choices(self) -> int:
        """Number of all choices.

        Note:
            Since derive mutable does not have its own search space, the number
            of choices will always be `1`.

        Returns:
            int: Number of choices.
        """
        return 1

    @property
    def current_choice(self) -> CHOICE_TYPE:
        """Current choice of derived mutable."""
        return self.choice_fn()

    @current_choice.setter
    def current_choice(self, choice: CHOICE_TYPE) -> None:
        """Setter of current choice.

        Raises:
            RuntimeError: Error when `current_choice` of derived mutable
                is modified directly.
        """
        raise RuntimeError('Choice of drived mutable can not be set.')

    @property
    def current_mask(self) -> Tensor:
        """Current mask of derived mutable."""
        if self.mask_fn is None:
            raise RuntimeError(
                '`mask_fn` must be set before access `current_mask`.')
        return self.mask_fn()

    @current_mask.setter
    def current_mask(self, mask: Tensor) -> None:
        """Setter of current mask.

        Raises:
            RuntimeError: Error when `current_mask` of derived mutable
                is modified directly.
        """
        raise RuntimeError('Mask of drived mutable can not be set.')

    @staticmethod
    def _trace_source_mutables_from_closure(
            closure: Callable) -> Set[BaseMutable]:
        """Trace source mutables from closure."""
        source_mutables: Set[BaseMutable] = set()

        def add_mutables_dfs(
                mutable: Union[Iterable, BaseMutable, Dict]) -> None:
            nonlocal source_mutables
            if isinstance(mutable, BaseMutable):
                if isinstance(mutable, DerivedMutable):
                    source_mutables |= mutable.source_mutables
                else:
                    source_mutables.add(mutable)
            # dict is also iterable, should parse first
            elif isinstance(mutable, dict):
                add_mutables_dfs(mutable.values())
                add_mutables_dfs(mutable.keys())
            elif isinstance(mutable, Iterable):
                for m in mutable:
                    add_mutables_dfs(m)

        noncolcal_pars = inspect.getclosurevars(closure).nonlocals
        add_mutables_dfs(noncolcal_pars.values())

        return source_mutables

    def _trace_source_mutables(self) -> Set[BaseMutable]:
        """Trace source mutables."""
        source_mutables = self._trace_source_mutables_from_closure(
            self.choice_fn)
        if self.mask_fn is not None:
            source_mutables |= self._trace_source_mutables_from_closure(
                self.mask_fn)

        return source_mutables

    @staticmethod
    def is_source_mutable(mutable: object) -> bool:
        """Judge whether an object is source mutable(not derived mutable).

        Args:
            mutable (object): An object.

        Returns:
            bool: Indicate whether the object is source mutable or not.
        """
        return isinstance(mutable, BaseMutable) and \
            not isinstance(mutable, DerivedMutable)

    # TODO
    # should be __str__? but can not provide info when debug
    def __repr__(self) -> str:  # pragma: no cover
        s = f'{self.__class__.__name__}('
        s += f'current_choice={self.current_choice}, '
        if self.mask_fn is not None:
            s += f'activated_channels={self.current_mask.sum().item()}, '
        s += f'source_mutables={self.source_mutables}, '
        s += f'is_fixed={self.is_fixed})'

        return s



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn

from mmrazor.registry import MODELS


@MODELS.register_module()
class ABLoss(nn.Module):
    """Activation Boundaries Loss.

    Paper: Knowledge Transfer via Distillation of Activation Boundaries
    Formed by Hidden Neurons, AAAI2019. https://arxiv.org/pdf/1811.03233.pdf

    Modified from: https://github.com/facebookresearch/AlphaNet

    Args:
        loss_weight (float): Weight of loss. Defaults to 1.0.
        margin (float): Relaxation for training stability. Defaults to 1.0.
    """

    def __init__(
        self,
        loss_weight: float = 1.0,
        margin: float = 1.0,
    ) -> None:
        super().__init__()
        self.loss_weight = loss_weight
        self.margin = margin

    def forward(
        self,
        s_feature: torch.Tensor,
        t_feature: torch.Tensor,
    ) -> torch.Tensor:
        """ABLoss forward function.

        Args:
            s_features (torch.Tensor): Student featuremap.
            t_features (torch.Tensor): Teacher featuremap.
        """
        batch_size = s_feature.shape[0]
        loss = self.criterion_alternative_l2(s_feature, t_feature)
        loss = loss / batch_size / 1000 * 3
        return self.loss_weight * loss

    def criterion_alternative_l2(
        self,
        source: torch.Tensor,
        target: torch.Tensor,
    ) -> torch.Tensor:
        """Piecewise differentiable loss approximating the activation
        boundaries loss.

        Guide the student learns a separating boundary between activation
        region and deactivation region formed by each neuron in the teacher.

        Args:
            source (torch.Tensor): Student featuremap.
            target (torch.Tensor): Teacher featuremap.
        """
        loss = ((source + self.margin)**2 * ((source > -self.margin) &
                                             (target <= 0)).float() +
                (source - self.margin)**2 * ((source <= self.margin) &
                                             (target > 0)).float())
        return torch.abs(loss).sum()



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional

import torch

from mmrazor.registry import MODELS
from .base_connector import BaseConnector


@MODELS.register_module()
class OFDTeacherConnector(BaseConnector):
    """Connector designed for ``OverhaulFeatureDistillation``

    Args:
        init_cfg (Optional[Dict], optional): Initialization config dict.
            Defaults to None.
    """

    def __init__(self, init_cfg: Optional[Dict] = None) -> None:
        super().__init__(init_cfg)
        self.margin: torch.Tensor = None

    def init_margin(self, margin: torch.Tensor) -> None:
        """Initializing margin, will be called by
        ``OverhaulFeatureDistillation``.

        Args:
            margin (torch.Tensor): margin
        """
        self.margin = margin

    def forward_train(self, feature: torch.Tensor) -> torch.Tensor:
        """forward func for training."""
        assert self.margin is not None, (
            'margin must be initialized before training.')
        self.margin = self.margin.to(feature.device)
        feature = torch.max(feature.detach(), self.margin)
        return feature



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Tuple, Union


def _addindent(s_, numSpaces):
    s = s_.split('\n')
    # don't do anything for single-line stuff
    if len(s) == 1:
        return s_
    first = s.pop(0)
    s = [(numSpaces * ' ') + line for line in s]
    s = '\n'.join(s)
    s = first + '\n' + s
    return s


def _merge_node_parents(node2parents, _node2parents):
    for node, parents in _node2parents.items():
        if node in node2parents:
            cur_parents = node2parents[node]
            new_parents_set = set(cur_parents + parents)
            new_parents = list(new_parents_set)
            node2parents[node] = new_parents
        else:
            node2parents[node] = parents


class PathNode:
    """``Node`` is the data structure that represents individual instances
    within a ``Path``. It corresponds to a module or an operation such as
    concatenation in the model.

    Args:
        name (str): Unique identifier of a node.
    """

    def __init__(self, name: str) -> None:
        self._name = name

    def get_module_names(self) -> List:
        return [self.name]

    @property
    def name(self) -> str:
        """Get the name of current node."""
        return self._name

    def _get_class_name(self):
        return self.__class__.__name__

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return f'{self._get_class_name()}(\'{self.name}\')'


class PathConvNode(PathNode):
    """A `ConvNode` corresponds to a Conv module in the original model."""
    pass


class PathDepthWiseConvNode(PathNode):
    """A `DepthWiseConvNode` corresponds to a depth-wise conv module in the
    original model."""
    pass


class PathNormNode(PathNode):
    """A `NormNode` corresponds to a normalization module in the original
    model."""
    pass


class PathLinearNode(PathNode):
    """A `LinearNode` corresponds to a linear module in the original model."""
    pass


class Path:
    """``Path`` is the data structure that represents a list of ``Node`` traced
    by a tracer.

    Args:
        nodes(:obj:`Node` or List[:obj:`Node`], optional): Nodes in a path.
            Default to None.
    """

    def __init__(self,
                 nodes: Optional[Union[PathNode, List[PathNode]]] = None):
        self._nodes: List[PathNode] = list()
        if nodes is not None:
            if isinstance(nodes, PathNode):
                nodes = [nodes]
            assert isinstance(nodes, (list, tuple))
            for node in nodes:
                assert isinstance(node, PathNode)
                self._nodes.append(node)

    def get_root_names(self) -> List[str]:
        """Get the name of the first node in a path."""
        return self._nodes[0].get_module_names()

    def find_nodes_parents(self,
                           target_nodes: Tuple,
                           non_pass: Optional[Tuple] = None) -> Dict:
        """Find the parents of a specific node.

        Args:
            target_nodes (Tuple): Find the parents of nodes whose types
                are one of `target_nodes`.
            non_pass (Tuple): Ancestor nodes whose types are one of
                `non_pass` are the parents of a specific node. Default to None.
        """
        node2parents: Dict[str, List[PathNode]] = dict()
        for i, node in enumerate(self._nodes):
            if isinstance(node, PathConcatNode):
                _node2parents: Dict[str,
                                    List[PathNode]] = node.find_nodes_parents(
                                        target_nodes, non_pass)
                _merge_node_parents(node2parents, _node2parents)
                continue

            if isinstance(node, target_nodes):
                parents = list()
                for behind_node in self._nodes[i + 1:]:
                    if non_pass is None or isinstance(behind_node, non_pass):
                        parents.append(behind_node)
                        break
                _node2parents = {node.name: parents}
                _merge_node_parents(node2parents, _node2parents)
        return node2parents

    @property
    def nodes(self) -> List:
        """Return a list of nodes in the current path."""
        return self._nodes

    def append(self, x: PathNode) -> None:
        """Add a node to the end of the current path."""
        assert isinstance(x, PathNode)
        self._nodes.append(x)

    def pop(self, *args, **kwargs):
        """Temoves the node at the given index from the path and returns the
        removed node."""
        return self._nodes.pop(*args, **kwargs)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.nodes == other.nodes
        else:
            return False

    def __len__(self):
        return len(self._nodes)

    def __getitem__(self, item):
        return self._nodes[item]

    def __iter__(self):
        for node in self._nodes:
            yield node

    def _get_class_name(self) -> str:
        """Get the name of the current class."""
        return self.__class__.__name__

    def __repr__(self):
        child_lines = []
        for node in self._nodes:
            node_str = repr(node)
            node_str = _addindent(node_str, 2)
            child_lines.append(node_str)
        lines = child_lines

        main_str = self._get_class_name() + '('
        if lines:
            main_str += '\n  ' + ',\n  '.join(lines) + '\n'
        main_str += ')'
        return main_str


class PathList:
    """``PathList`` is the data structure that represents a list of ``Path``
    traced by a tracer.

    Args:
        paths(:obj:`Path` or List[:obj:`Path`], optional): A list of `Path`.
            Default to None.
    """

    def __init__(self, paths: Optional[Union[Path, List[Path]]] = None):
        self._paths = list()
        if paths is not None:
            if isinstance(paths, Path):
                paths = [paths]
            assert isinstance(paths, (list, tuple))
            for path in paths:
                assert isinstance(path, Path)
                self._paths.append(path)

    def get_root_names(self) -> List[str]:
        """Get the root node of all the paths in `PathList`."""
        root_name_list = [path.get_root_names() for path in self._paths]
        for root_names in root_name_list[1:]:
            assert root_names == root_name_list[0], \
                f'If the input of a module is a concatenation of several ' \
                f'modules\' outputs, we can use `get_root_names` to get the' \
                f' names of these modules. As `get_root_names` is only used' \
                f' in this case, each element in `root_name_list` should be' \
                f' the same. Got root_name_list = {root_name_list}'
        return self._paths[0].get_root_names()

    def find_nodes_parents(self,
                           target_nodes: Tuple,
                           non_pass: Optional[Tuple] = None):
        """Find the parents of a specific node.

        Args:
            target_nodes (Tuple): Find the parents of nodes whose types
                are one of `target_nodes`.
            non_pass (Tuple): Ancestor nodes whose types are one of
                `non_pass` are the parents of a specific node. Default to None.
        """
        node2parents: Dict[str, List[PathNode]] = dict()
        for p in self._paths:
            _node2parents = p.find_nodes_parents(target_nodes, non_pass)
            _merge_node_parents(node2parents, _node2parents)
        return node2parents

    def append(self, x: Path) -> None:
        """Add a path to the end of the current PathList."""
        assert isinstance(x, Path)
        self._paths.append(x)

    @property
    def paths(self):
        """Return all paths in the current PathList."""
        return self._paths

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.paths == other.paths
        else:
            return False

    def __len__(self):
        return len(self._paths)

    def __getitem__(self, item):
        return self._paths[item]

    def __iter__(self):
        for path in self._paths:
            yield path

    def _get_class_name(self) -> str:
        """Get the name of the current class."""
        return self.__class__.__name__

    def __repr__(self):
        child_lines = []
        for node in self._paths:
            node_str = repr(node)
            node_str = _addindent(node_str, 2)
            child_lines.append(node_str)
        lines = child_lines

        main_str = self._get_class_name() + '('
        if lines:
            main_str += '\n  ' + ',\n  '.join(lines) + '\n'
        main_str += ')'
        return main_str


class PathConcatNode(PathNode):
    """``ConcatNode`` is the data structure that represents the concatenation
    operation in a model.

    Args:
        name (str): Unique identifier of a `ConcatNode`.
        path_lists (List[PathList]): Several nodes are concatenated and each
            node is the root node of all the paths in a `PathList`
            (one of `path_lists`).
    """

    def __init__(self, name: str, path_lists: List[PathList]):
        super().__init__(name)
        self._path_lists = list()
        for path_list in path_lists:
            assert isinstance(path_list, PathList)
            self._path_lists.append(path_list)

    def get_module_names(self) -> List[str]:
        """Several nodes are concatenated.

        Get the names of these nodes.
        """
        module_names = list()
        for path_list in self._path_lists:
            module_names.extend(path_list.get_root_names())
        return module_names

    def find_nodes_parents(self,
                           target_nodes: Tuple,
                           non_pass: Optional[Tuple] = None):
        """Find the parents of a specific node.

        Args:
            target_nodes (Tuple): Find the parents of nodes whose types
                are one of `target_nodes`.
            non_pass (Tuple): Ancestor nodes whose types are one of
                `non_pass` are the parents of a specific node. Default to None.
        """
        node2parents: Dict[str, List[PathNode]] = dict()
        for p in self._path_lists:
            _node2parents = p.find_nodes_parents(target_nodes, non_pass)
            _merge_node_parents(node2parents, _node2parents)
        return node2parents

    @property
    def path_lists(self) -> List[PathList]:
        """Return all the path_list."""
        return self._path_lists

    def __len__(self):
        return len(self._path_lists)

    def __getitem__(self, item):
        return self._path_lists[item]

    def __iter__(self):
        for path_list in self._path_lists:
            yield path_list

    def _get_class_name(self) -> str:
        """Get the name of the current class."""
        return self.__class__.__name__

    def __repr__(self):
        child_lines = []
        for node in self._path_lists:
            node_str = repr(node)
            node_str = _addindent(node_str, 2)
            child_lines.append(node_str)
        lines = child_lines

        main_str = self._get_class_name() + '('
        if lines:
            main_str += '\n  ' + ',\n  '.join(lines) + '\n'
        main_str += ')'
        return main_str



_base_ = ['./cwd_fpn_frcnn_r101_frcnn_r50_1x_coco.py']

model = dict(
    architecture=dict(
        cfg_path='mmdet::retinanet/retinanet_r50_fpn_1x_coco.py',
        pretrained=False),
    teacher=dict(
        cfg_path='mmdet::retinanet/retinanet_r101_fpn_2x_coco.py',
        pretrained=True))

# optimizer
optim_wrapper = dict(
    optimizer=dict(type='SGD', lr=0.01, momentum=0.9, weight_decay=0.0001))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional

import torch
from mmengine.structures import BaseDataElement

from mmrazor.models.utils import add_prefix
from mmrazor.registry import MODELS
from ...base import LossResults
from .single_teacher_distill import SingleTeacherDistill


@MODELS.register_module()
class FpnTeacherDistill(SingleTeacherDistill):
    """``FpnTeacherDistill`` means teacher only execute backbone and neck.

    If the intermediate results required for distill algorithm are generated by
    the backbone and neck parts, using ``FpnTeacherDistill`` can speed up
    training.
    """

    def loss(
        self,
        batch_inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
    ) -> LossResults:
        """Calculate losses from a batch of inputs and data samples."""

        losses = dict()
        # If the `override_data` of a delivery is False, the delivery will
        # record the origin data.
        self.distiller.set_deliveries_override(False)
        if self.teacher_trainable:
            # Unlike ``SingleTeacherDistill``, teacher will only execute
            # back + neck, not head, so there will be no loss.
            with self.distiller.teacher_recorders, self.distiller.deliveries:
                _ = self.teacher.extract_feat(batch_inputs)
        else:
            with self.distiller.teacher_recorders, self.distiller.deliveries:
                with torch.no_grad():
                    _ = self.teacher(batch_inputs, data_samples, mode='loss')

        # If the `override_data` of a delivery is True, the delivery will
        # override the origin data with the recorded data.
        self.distiller.set_deliveries_override(True)
        with self.distiller.student_recorders, self.distiller.deliveries:
            student_losses = self.student(
                batch_inputs, data_samples, mode='loss')
        losses.update(add_prefix(student_losses, 'student'))

        # Automatically compute distill losses based on `loss_forward_mappings`
        # The required data already exists in the recorders.
        distill_losses = self.distiller.compute_distill_losses()
        losses.update(add_prefix(distill_losses, 'distill'))

        return losses



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine.dist import get_dist_info

from mmrazor.registry import MODELS
from ..architectures.ops import GatherTensors


class DAFLLoss(nn.Module):
    """Base class for DAFL losses.

    paper link: https://arxiv.org/pdf/1904.01186.pdf

    Args:
        loss_weight (float): Weight of the loss. Defaults to 1.0.
    """

    def __init__(self, loss_weight=1.0) -> None:
        super().__init__()
        self.loss_weight = loss_weight

    def forward(self, preds_T: torch.Tensor) -> torch.Tensor:
        """Forward function for the DAFLLoss.

        Args:
            preds_T (torch.Tensor): The predictions of teacher.
        """
        return self.loss_weight * self.forward_train(preds_T)

    def forward_train(self, preds_T: torch.Tensor) -> torch.Tensor:
        """Forward function during training.

        Args:
            preds_T (torch.Tensor): The predictions of teacher.
        """
        raise NotImplementedError


@MODELS.register_module()
class OnehotLikeLoss(DAFLLoss):
    """The loss function for measuring the one-hot-likeness of the target
    logits."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def forward_train(self, preds_T: torch.Tensor) -> torch.Tensor:
        """Forward function in training for the OnehotLikeLoss.

        Args:
            preds_T (torch.Tensor): The predictions of teacher.
        """
        fake_label = preds_T.data.max(1)[1]
        return F.cross_entropy(preds_T, fake_label)


@MODELS.register_module()
class InformationEntropyLoss(DAFLLoss):
    """The loss function for measuring the class balance of the target logits.

    Args:
        gather (bool, optional): The switch controlling whether
            collecting tensors from multiple gpus. Defaults to True.
    """

    def __init__(self, gather=True, **kwargs) -> None:
        super().__init__(**kwargs)
        self.gather = gather
        _, self.world_size = get_dist_info()

    def forward_train(self, preds_T: torch.Tensor) -> torch.Tensor:
        """Forward function in training for the InformationEntropyLoss.

        Args:
            preds_T (torch.Tensor): The predictions of teacher.
        """
        # Gather predictions from all GPUS to calibrate the loss function.
        if self.gather and self.world_size > 1:
            preds_T = torch.cat(GatherTensors.apply(preds_T), dim=0)
        class_prob = F.softmax(preds_T, dim=1).mean(dim=0)
        info_entropy = class_prob * torch.log10(class_prob)
        return info_entropy.sum()


@MODELS.register_module()
class ActivationLoss(nn.Module):
    """The loss function for measuring the activation of the target featuremap.
    It is negative of the norm of the target featuremap.

    Args:
        loss_weight (float): Weight of the loss. Defaults to 1.0.
        norm_type (str, optional):The type of the norm. Defaults to 'abs'.
    """

    def __init__(self, loss_weight=1.0, norm_type='abs') -> None:
        super().__init__()
        self.loss_weight = loss_weight
        assert norm_type in ['norm', 'abs'], \
            '"norm_type" must be "norm" or "abs"'
        self.norm_type = norm_type

        if self.norm_type == 'norm':
            self.norm_fn = lambda x: -x.norm()
        elif self.norm_type == 'abs':
            self.norm_fn = lambda x: -x.abs().mean()

    def forward(self, feat_T: torch.Tensor) -> torch.Tensor:
        """Forward function for the ActivationLoss.

        Args:
            feat_T (torch.Tensor): The featuremap of teacher.
        """
        return self.loss_weight * self.forward_train(feat_T)

    def forward_train(self, feat_T: torch.Tensor) -> torch.Tensor:
        """Forward function in training for the ActivationLoss.

        Args:
            feat_T (torch.Tensor): The featuremap of teacher.
        """
        feat_T = feat_T.view(feat_T.size(0), -1)
        return self.norm_fn(feat_T)



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABC, abstractmethod
from typing import Dict, Generic, Optional, Type, TypeVar

from mmengine.model import BaseModule
from torch.nn import Module

from ..mutables.base_mutable import BaseMutable

MUTABLE_TYPE = TypeVar('MUTABLE_TYPE', bound=BaseMutable)


class BaseMutator(ABC, BaseModule, Generic[MUTABLE_TYPE]):
    """The base class for mutator.

    Mutator is mainly used for subnet management, it usually provides functions
    such as sampling and setting of subnets.

    All subclasses should implement the following APIs:

    - ``prepare_from_supernet()``
    - ``search_space``

    Args:
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(self, init_cfg: Optional[Dict] = None) -> None:
        super().__init__(init_cfg=init_cfg)

    @abstractmethod
    def prepare_from_supernet(self, supernet: Module) -> None:
        """Do some necessary preparations with supernet.

        Args:
            supernet (:obj:`torch.nn.Module`): The supernet to be searched
                in your algorithm.
        """

    @property
    @abstractmethod
    def search_groups(self) -> Dict:
        """Search group of the supernet.

        Note:
            Search group is different from search space. The key of search
            group is called ``group_id``, and the value is corresponding
            searchable modules. The searchable modules will have the same
            search space if they are in the same group.

        Returns:
            dict: Search group.
        """

    @property
    @abstractmethod
    def mutable_class_type(self) -> Type[MUTABLE_TYPE]:
        """Corresponding mutable class type."""
        pass



# Copyright (c) OpenMMLab. All rights reserved.
import random
from typing import Any, Dict, List, Optional, Tuple, Union

from mmrazor.registry import MODELS
from ..base_mutable import BaseMutable
from ..derived_mutable import DerivedMethodMixin, DerivedMutable


@MODELS.register_module()
class MutableValue(BaseMutable[Any, Dict], DerivedMethodMixin):
    """Base class for mutable value.

    A mutable value is actually a mutable that adds some functionality to a
    list containing objects of the same type.

    Args:
        value_list (list): List of value, each value must have the same type.
        default_value (any, optional): Default value, must be one in
            `value_list`. Default to None.
        alias (str, optional): alias of the `MUTABLE`.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.
    """

    def __init__(self,
                 value_list: List[Any],
                 default_value: Optional[Any] = None,
                 alias: Optional[str] = None,
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(alias, init_cfg)

        self._check_is_same_type(value_list)
        self._value_list = value_list

        if default_value is None:
            default_value = value_list[0]
        self.current_choice = default_value

    @staticmethod
    def _check_is_same_type(value_list: List[Any]) -> None:
        """Check whether value in `value_list` has the same type."""
        if len(value_list) == 1:
            return

        for i in range(1, len(value_list)):
            is_same_type = type(value_list[i - 1]) is \
                type(value_list[i])  # noqa: E721
            if not is_same_type:
                raise TypeError(
                    'All elements in `value_list` must have same '
                    f'type, but both types {type(value_list[i-1])} '
                    f'and type {type(value_list[i])} exist.')

    @property
    def choices(self) -> List[Any]:
        """List of choices."""
        return self._value_list

    def fix_chosen(self, chosen: Dict[str, Any]) -> None:
        """Fix mutable value with subnet config.

        Args:
            chosen (dict): the information of chosen.
        """
        if self.is_fixed:
            raise RuntimeError('MutableValue can not be fixed twice')

        all_choices = chosen['all_choices']
        current_choice = chosen['current_choice']

        assert all_choices == self.choices, \
            f'Expect choices to be: {self.choices}, but got: {all_choices}'
        assert current_choice in self.choices

        self.current_choice = current_choice
        self.is_fixed = True

    def dump_chosen(self) -> Dict[str, Any]:
        """Dump information of chosen.

        Returns:
            Dict[str, Any]: Dumped information.
        """
        return dict(
            current_choice=self.current_choice, all_choices=self.choices)

    @property
    def num_choices(self) -> int:
        """Number of all choices.

        Returns:
            int: Number of choices.
        """
        return len(self.choices)

    @property
    def current_choice(self) -> Optional[Any]:
        """Current choice of mutable value."""
        return self._current_choice

    @current_choice.setter
    def current_choice(self, choice: Any) -> Any:
        """Setter of current choice."""
        if choice not in self.choices:
            raise ValueError(f'Expected choice in: {self.choices}, '
                             f'but got: {choice}')

        self._current_choice = choice

    def __rmul__(self, other) -> DerivedMutable:
        """Please refer to method :func:`__mul__`."""
        return self * other

    def __mul__(self, other: int) -> DerivedMutable:
        """Overload `*` operator.

        Args:
            other (int): Expand ratio.

        Returns:
            DerivedMutable: Derived expand mutable.
        """
        if isinstance(other, int):
            return self.derive_expand_mutable(other)

        raise TypeError(f'Unsupported type {type(other)} for mul!')

    def __floordiv__(self, other: Union[int, Tuple[int,
                                                   int]]) -> DerivedMutable:
        """Overload `//` operator.

        Args:
            other: (int, tuple): divide ratio for int or
                (divide ratio, divisor) for tuple.

        Returns:
            DerivedMutable: Derived divide mutable.
        """
        if isinstance(other, int):
            return self.derive_divide_mutable(other)
        if isinstance(other, tuple):
            assert len(other) == 2
            return self.derive_divide_mutable(*other)

        raise TypeError(f'Unsupported type {type(other)} for div!')

    def __repr__(self) -> str:
        s = self.__class__.__name__
        s += f'(value_list={self._value_list}, '
        s += f'current_choice={self.current_choice})'

        return s


# TODO
# 1. use comparable for type hint
# 2. use mixin
@MODELS.register_module()
class OneShotMutableValue(MutableValue):
    """Class for one-shot mutable value.

    one-shot mutable value provides `sample_choice` method and `min_choice`,
    `max_choice` properties on the top of mutable value.

    Args:
        value_list (list): List of value, each value must have the same type.
        default_value (any, optional): Default value, must be one in
            `value_list`. Default to None.
        alias (str, optional): alias of the `MUTABLE`.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.
    """

    def __init__(self,
                 value_list: List[Any],
                 default_value: Optional[Any] = None,
                 alias: Optional[str] = None,
                 init_cfg: Optional[Dict] = None) -> None:
        value_list = sorted(value_list)
        # set default value as max value
        if default_value is None:
            default_value = value_list[-1]

        super().__init__(
            value_list=value_list,
            default_value=default_value,
            alias=alias,
            init_cfg=init_cfg)

    def sample_choice(self) -> Any:
        """Random sampling from choices.

        Returns:
            Any: Selected choice.
        """
        return random.choice(self.choices)

    @property
    def max_choice(self) -> Any:
        """Max choice of all choices.

        Returns:
            Any: Max choice.
        """
        return self.choices[-1]

    @property
    def min_choice(self) -> Any:
        """Min choice of all choices.

        Returns:
            Any: Min choice.
        """
        return self.choices[0]

    def __mul__(self, other) -> DerivedMutable:
        """Overload `*` operator.

        Args:
            other (int, OneShotMutableChannel): Expand ratio or
                OneShotMutableChannel.

        Returns:
            DerivedMutable: Derived expand mutable.
        """
        from ..mutable_channel import OneShotMutableChannel

        if isinstance(other, OneShotMutableChannel):
            return other * self

        return super().__mul__(other)



_base_ = [
    'mmcls::_base_/datasets/imagenet_bs32.py',
    'mmcls::_base_/schedules/imagenet_bs256.py',
    'mmcls::_base_/default_runtime.py'
]

teacher_ckpt = 'https://download.openmmlab.com/mmclassification/v0/resnet/resnet50_8xb32_in1k_20210831-ea4938fc.pth'  # noqa: E501
model = dict(
    _scope_='mmrazor',
    type='SingleTeacherDistill',
    data_preprocessor=dict(
        type='ImgDataPreprocessor',
        # RGB format normalization parameters
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        # convert image from BGR to RGB
        bgr_to_rgb=True),
    architecture=dict(
        cfg_path='mmcls::resnet/resnet18_8xb32_in1k.py', pretrained=False),
    teacher=dict(
        cfg_path='mmcls::resnet/resnet50_8xb32_in1k.py', pretrained=True),
    teacher_ckpt=teacher_ckpt,
    calculate_student_loss=False,
    distiller=dict(
        type='ConfigurableDistiller',
        student_recorders=dict(
            bb_s4=dict(type='ModuleOutputs', source='backbone.layer4.1.conv2'),
            bb_s3=dict(type='ModuleOutputs', source='backbone.layer3.1.conv2'),
            bb_s2=dict(type='ModuleOutputs', source='backbone.layer2.1.conv2'),
            bb_s1=dict(type='ModuleOutputs',
                       source='backbone.layer1.1.conv2')),
        teacher_recorders=dict(
            bb_s4=dict(type='ModuleOutputs', source='backbone.layer4.2.conv3'),
            bb_s3=dict(type='ModuleOutputs', source='backbone.layer3.5.conv3'),
            bb_s2=dict(type='ModuleOutputs', source='backbone.layer2.3.conv3'),
            bb_s1=dict(type='ModuleOutputs',
                       source='backbone.layer1.2.conv3')),
        distill_losses=dict(
            loss_s4=dict(type='ABLoss', loss_weight=1.0),
            loss_s3=dict(type='ABLoss', loss_weight=0.5),
            loss_s2=dict(type='ABLoss', loss_weight=0.25),
            loss_s1=dict(type='ABLoss', loss_weight=0.125)),
        connectors=dict(
            loss_s4_sfeat=dict(
                type='ConvModuleConncetor',
                in_channel=512,
                out_channel=2048,
                norm_cfg=dict(type='BN'),
                act_cfg=None),
            loss_s3_sfeat=dict(
                type='ConvModuleConncetor',
                in_channel=256,
                out_channel=1024,
                norm_cfg=dict(type='BN'),
                act_cfg=None),
            loss_s2_sfeat=dict(
                type='ConvModuleConncetor',
                in_channel=128,
                out_channel=512,
                norm_cfg=dict(type='BN'),
                act_cfg=None),
            loss_s1_sfeat=dict(
                type='ConvModuleConncetor',
                in_channel=64,
                out_channel=256,
                norm_cfg=dict(type='BN'),
                act_cfg=None)),
        loss_forward_mappings=dict(
            loss_s4=dict(
                s_feature=dict(
                    from_student=True,
                    recorder='bb_s4',
                    connector='loss_s4_sfeat'),
                t_feature=dict(from_student=False, recorder='bb_s4')),
            loss_s3=dict(
                s_feature=dict(
                    from_student=True,
                    recorder='bb_s3',
                    connector='loss_s3_sfeat'),
                t_feature=dict(from_student=False, recorder='bb_s3')),
            loss_s2=dict(
                s_feature=dict(
                    from_student=True,
                    recorder='bb_s2',
                    connector='loss_s2_sfeat'),
                t_feature=dict(from_student=False, recorder='bb_s2')),
            loss_s1=dict(
                s_feature=dict(
                    from_student=True,
                    recorder='bb_s1',
                    connector='loss_s1_sfeat'),
                t_feature=dict(from_student=False, recorder='bb_s1')))))

find_unused_parameters = True

train_cfg = dict(by_epoch=True, max_epochs=20, val_interval=1)
val_cfg = dict(_delete_=True, type='mmrazor.SingleTeacherDistillValLoop')



_base_ = [
    './factor-transfer_backbone_resnet50_resnet18_8xb16_cifar10_pretrain.py'
]

train_cfg = dict(by_epoch=True, max_epochs=200, val_interval=1)

model = dict(
    calculate_student_loss=True,
    student_trainable=True,
    distiller=dict(
        distill_losses=dict(loss_s4=dict(type='FTLoss', loss_weight=1.0)),
        connectors=dict(loss_s4_tfeat=dict(phase='train')),
        loss_forward_mappings=dict(
            _delete_=True,
            loss_s4=dict(
                s_feature=dict(
                    from_student=True,
                    recorder='bb_s4',
                    connector='loss_s4_sfeat'),
                t_feature=dict(
                    from_student=False,
                    recorder='bb_s4',
                    connector='loss_s4_tfeat'),
            ))),
    init_cfg=dict(
        type='Pretrained',
        checkpoint=  # noqa: E251
        'https://download.openmmlab.com/mmrazor/v1/factor_transfer/factor-transfer_backbone_resnet50_resnet18_8xb16_cifar10_pretrain_20220831_173259-ebdb09e2.pth'  # noqa: E501
    ))



# Copyright (c) OpenMMLab. All rights reserved.
from mmrazor.registry import TASK_UTILS
from ..flops_params_counter import get_model_parameters_number
from .base_counter import BaseCounter


@TASK_UTILS.register_module()
class ConvTranspose2dCounter(BaseCounter):
    """FLOPs/params counter for Decov module series."""

    @staticmethod
    def add_count_hook(module, input, output):
        # Can have multiple inputs, getting the first one
        input = input[0]

        batch_size = input.shape[0]
        input_height, input_width = input.shape[2:]

        # TODO: use more common representation
        kernel_height, kernel_width = module.kernel_size
        in_channels = module.in_channels
        out_channels = module.out_channels
        groups = module.groups

        filters_per_channel = out_channels // groups
        conv_per_position_flops = (
            kernel_height * kernel_width * in_channels * filters_per_channel)

        active_elements_count = batch_size * input_height * input_width
        overall_conv_flops = conv_per_position_flops * active_elements_count
        bias_flops = 0
        if module.bias is not None:
            output_height, output_width = output.shape[2:]
            bias_flops = out_channels * batch_size * output_height * output_height  # noqa: E501
        overall_flops = overall_conv_flops + bias_flops

        module.__flops__ += int(overall_flops)
        module.__params__ += get_model_parameters_number(module)



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Dict, Optional

from torch import nn

from mmrazor.registry import TASK_UTILS
from .base_recorder import BaseRecorder


class RecorderManager:
    """Various types recorders' manager. The ``RecorderManager`` also is a
    context manager, managing various types of Recorder. When entering the
    ``RecorderManager``, all recorders managed by it will be started.

    Note:
        The recorders will be initialized in the ``RecorderManager`` by
        default. If you want to just use a recorder without the
        ``RecorderManager``, you need to initialize it first.

    Args:
        recorders (dict, optional): All recorders' config.


    Examples:
        >>> # Below code in toy_module.py
        >>> import random
        >>> class Toy():
        ...     def toy_func(self):
        ...         return random.randint(0, 1000)

        >>> # Below code in main.py
        >>> from torch import nn
        >>> from toy_module import Toy

        >>> class ToyModel(nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.conv1 = nn.Conv2d(1,1,1)
        ...         self.conv2 = nn.Conv2d(1,1,1)
        ...         self.toy = Toy()
        ...     def forward(self, x):
        ...         return self.conv2(self.conv1(x)) + self.toy.toy_func()

        >>> model = ToyModel()
        >>> [ name for name,_ in model.named_modules() ]
        ['conv1', 'conv2']

        >>> conv1_rec = ModuleOutputsRecorder('conv1')
        >>> conv2_rec = ModuleOutputsRecorder('conv2')
        >>> func_rec = MethodOutputsRecorder('toy_module.Toy.toy_func')
        >>> manager = RecorderManager(
        ...             {'conv1_rec': conv1_rec ,
        ...              'conv2_rec': conv2_rec,
        ...              'func_rec': func_rec})
        >>> manager.initialize(model)

        >>> with manager:
        ...     res = model(torch.ones(1,1,1,1))
        >>> res
        tensor([[[[22.9534]]]])

        >>> conv2_data = manager.get_recorder('conv2_rec').get_record_data()
        >>> conv2_data
        tensor([[[[0.9534]]]])

        >>> func_data = manager.get_recorder('func_rec').get_record_data()
        >>> func_data
        22

        >>> res.sum() == (conv2_data + func_data).sum()
        True
    """

    def __init__(self, recorders: Optional[Dict] = None) -> None:

        self._recorders: Dict[str, BaseRecorder] = dict()
        if recorders:
            for name, cfg in recorders.items():
                recorder_cfg = copy.deepcopy(cfg)
                recorder_type = cfg['type']
                recorder_type_ = recorder_type + 'Recorder'

                recorder_cfg['type'] = recorder_type_
                recorder = TASK_UTILS.build(recorder_cfg)

                self._recorders[name] = recorder

    @property
    def recorders(self) -> Dict[str, BaseRecorder]:
        """dict: all recorders."""
        return self._recorders

    def get_recorder(self, recorder: str) -> BaseRecorder:
        """Get the corresponding recorder according to the name."""
        return self.recorders[recorder]

    def initialize(self, model: nn.Module):
        """Init all recorders.

        Args:
            model (nn.Module): The model which need to record intermediate
                results.
        """
        for recorder in self.recorders.values():
            recorder.initialize(model)

    def __enter__(self):
        """Enter the context manager."""
        for recorder in self.recorders.values():
            recorder.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager."""
        for recorder in self.recorders.values():
            recorder.__exit__(exc_type, exc_value, traceback)



# Copyright (c) OpenMMLab. All rights reserved.
import copy
import os
from pathlib import Path
from typing import Dict, List, Optional, Union

import torch
from mmengine import fileio
from mmengine.model import BaseModel, MMDistributedDataParallel
from mmengine.optim import OptimWrapper
from mmengine.structures import BaseDataElement
from torch import nn

from mmrazor.models.mutators import SlimmableChannelMutator
from mmrazor.models.utils import (add_prefix,
                                  reinitialize_optim_wrapper_count_status)
from mmrazor.registry import MODEL_WRAPPERS, MODELS
from ..base import BaseAlgorithm

VALID_MUTATOR_TYPE = Union[SlimmableChannelMutator, Dict]
VALID_PATH_TYPE = Union[str, Path]
VALID_CHANNEL_CFG_PATH_TYPE = Union[VALID_PATH_TYPE, List[VALID_PATH_TYPE]]


@MODELS.register_module()
class SlimmableNetwork(BaseAlgorithm):
    """Slimmable Neural Networks.

    Please refer to paper
    [Slimmable Neural Networks](https://arxiv.org/abs/1812.08928) for details.

    Args:
        mutator (dict | :obj:`SlimmableChannelMutator`): The config of
            :class:`SlimmableChannelMutator` or built mutator.
        architecture (dict | :obj:`BaseModel`): The config of
            :class:`BaseModel` or built model.
        channel_cfg_paths (str | :obj:`Path` | list): Config of list of configs
            for channel of subnet(s) searched out. If there is only one
            channel_cfg, the supernet will be fixed.
        data_preprocessor (dict | :obj:`torch.nn.Module` | None): The
            pre-process config of :class:`BaseDataPreprocessor`.
            Defaults to None.
        init_cfg (dict | None): The weight initialized config for
            :class:`BaseModule`. Default to None.
    """

    def __init__(self,
                 mutator: VALID_MUTATOR_TYPE,
                 architecture: Union[BaseModel, Dict],
                 channel_cfg_paths: VALID_CHANNEL_CFG_PATH_TYPE,
                 data_preprocessor: Optional[Union[Dict, nn.Module]] = None,
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(architecture, data_preprocessor, init_cfg)

        if not isinstance(channel_cfg_paths, list):
            channel_cfg_paths = [channel_cfg_paths]
        self.num_subnet = len(channel_cfg_paths)

        channel_cfgs = self._load_and_merge_channel_cfgs(channel_cfg_paths)
        if isinstance(mutator, dict):
            assert mutator.get('channel_cfgs') is None, \
                '`channel_cfgs` should not be in channel config'
            mutator = copy.deepcopy(mutator)
            mutator['channel_cfgs'] = channel_cfgs

        self.mutator: SlimmableChannelMutator = self._build_mutator(mutator)
        self.mutator.prepare_from_supernet(self.architecture)

        # must after `prepare_from_supernet`
        if len(channel_cfg_paths) == 1:
            # Avoid circular import
            from mmrazor.structures import load_fix_subnet
            load_fix_subnet(self.architecture, channel_cfg_paths[0])
            self.is_deployed = True
        else:
            self.is_deployed = False

        # HACK
        # reinitialize count status of `OptimWrapper` since
        # `optim_wrapper.update_params` will be called multiple times
        # in our slimmable train step.
        self._optim_wrapper_count_status_reinitialized = False

    def _load_and_merge_channel_cfgs(
            self, channel_cfg_paths: List[VALID_PATH_TYPE]) -> Dict:
        """Load and merge channel config."""
        channel_cfgs = list()
        for channel_cfg_path in channel_cfg_paths:
            channel_cfg = fileio.load(channel_cfg_path)
            channel_cfgs.append(channel_cfg)

        return self.merge_channel_cfgs(channel_cfgs)

    @staticmethod
    def merge_channel_cfgs(channel_cfgs: List[Dict]) -> Dict:
        """Merge several channel configs."""
        merged_channel_cfg = dict()
        num_subnet = len(channel_cfgs)

        for module_name in channel_cfgs[0].keys():
            channels_per_layer = [
                channel_cfgs[idx][module_name] for idx in range(num_subnet)
            ]
            merged_channels_per_layer = dict()
            for key in channels_per_layer[0].keys():
                merged_channels = [
                    channels_per_layer[idx][key] for idx in range(num_subnet)
                ]
                merged_channels_per_layer[key] = merged_channels
            merged_channel_cfg[module_name] = merged_channels_per_layer

        return merged_channel_cfg

    def _build_mutator(self,
                       mutator: VALID_MUTATOR_TYPE) -> SlimmableChannelMutator:
        """build mutator."""
        if isinstance(mutator, dict):
            mutator = MODELS.build(mutator)
        if not isinstance(mutator, SlimmableChannelMutator):
            raise TypeError('mutator should be a `dict` or '
                            '`SlimmableChannelMutator` instance, but got '
                            f'{type(mutator)}')

        return mutator

    def train_step(self, data: List[dict],
                   optim_wrapper: OptimWrapper) -> Dict[str, torch.Tensor]:
        """Train step."""
        batch_inputs, data_samples = self.data_preprocessor(data, True)
        train_kwargs = dict(
            batch_inputs=batch_inputs,
            data_samples=data_samples,
            optim_wrapper=optim_wrapper)
        if self.is_deployed:
            return self._fixed_train_step(**train_kwargs)
        else:
            return self._slimmable_train_step(**train_kwargs)

    def _slimmable_train_step(
        self,
        batch_inputs: torch.Tensor,
        data_samples: List[BaseDataElement],
        optim_wrapper: OptimWrapper,
    ) -> Dict[str, torch.Tensor]:
        """Train step of Slimmable Network."""
        if not self._optim_wrapper_count_status_reinitialized:
            reinitialize_optim_wrapper_count_status(
                model=self,
                optim_wrapper=optim_wrapper,
                accumulative_counts=self.num_subnet)
            self._optim_wrapper_count_status_reinitialized = True
        total_losses = dict()

        for subnet_idx in range(self.num_subnet):
            self.mutator.switch_choices(subnet_idx)
            with optim_wrapper.optim_context(self):
                losses = self(batch_inputs, data_samples, mode='loss')
            parsed_losses, _ = self.parse_losses(losses)
            optim_wrapper.update_params(parsed_losses)

            total_losses.update(add_prefix(losses, f'subnet_{subnet_idx}'))

        return total_losses

    def _fixed_train_step(
        self,
        batch_inputs: torch.Tensor,
        data_samples: List[BaseDataElement],
        optim_wrapper: OptimWrapper,
    ) -> Dict[str, torch.Tensor]:
        """Train step of fixed network."""
        with optim_wrapper.optim_context(self):
            losses = self(batch_inputs, data_samples, mode='loss')
        parsed_losses, _ = self.parse_losses(losses)
        optim_wrapper.update_params(parsed_losses)

        return losses


@MODEL_WRAPPERS.register_module()
class SlimmableNetworkDDP(MMDistributedDataParallel):
    """DDP wrapper for Slimmable Neural Network."""

    def __init__(self,
                 *,
                 device_ids: Optional[Union[List, int, torch.device]] = None,
                 **kwargs) -> None:
        if device_ids is None:
            if os.environ.get('LOCAL_RANK') is not None:
                device_ids = [int(os.environ['LOCAL_RANK'])]
        super().__init__(device_ids=device_ids, **kwargs)

    def train_step(self, data: List[dict],
                   optim_wrapper: OptimWrapper) -> Dict[str, torch.Tensor]:
        """Train step."""
        batch_inputs, data_samples = self.module.data_preprocessor(data, True)
        train_kwargs = dict(
            batch_inputs=batch_inputs,
            data_samples=data_samples,
            optim_wrapper=optim_wrapper)
        if self.module.is_deployed:
            return self._fixed_train_step(**train_kwargs)
        else:
            return self._slimmable_train_step(**train_kwargs)

    def _slimmable_train_step(
        self,
        batch_inputs: torch.Tensor,
        data_samples: List[BaseDataElement],
        optim_wrapper: OptimWrapper,
    ) -> Dict[str, torch.Tensor]:
        """Train step of Slimmable Network."""
        if not self._optim_wrapper_count_status_reinitialized:
            reinitialize_optim_wrapper_count_status(
                model=self,
                optim_wrapper=optim_wrapper,
                accumulative_counts=self.module.num_subnet)
            self._optim_wrapper_count_status_reinitialized = True
        total_losses = dict()

        for subnet_idx in range(self.module.num_subnet):
            self.module.mutator.switch_choices(subnet_idx)
            with optim_wrapper.optim_context(self):
                losses = self(batch_inputs, data_samples, mode='loss')
            parsed_losses, _ = self.module.parse_losses(losses)
            optim_wrapper.update_params(parsed_losses)

            total_losses.update(add_prefix(losses, f'subnet_{subnet_idx}'))

        return total_losses

    def _fixed_train_step(
        self,
        batch_inputs: torch.Tensor,
        data_samples: List[BaseDataElement],
        optim_wrapper: OptimWrapper,
    ) -> Dict[str, torch.Tensor]:
        """Train step of fixed network."""
        with optim_wrapper.optim_context(self):
            losses = self(batch_inputs, data_samples, mode='loss')
        parsed_losses, _ = self.module.parse_losses(losses)
        optim_wrapper.update_params(parsed_losses)

        return losses

    @property
    def _optim_wrapper_count_status_reinitialized(self) -> bool:
        return self.module._optim_wrapper_count_status_reinitialized

    @_optim_wrapper_count_status_reinitialized.setter
    def _optim_wrapper_count_status_reinitialized(self, val: bool) -> None:
        assert isinstance(val, bool)

        self.module._optim_wrapper_count_status_reinitialized = val



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Union

import torch
from mmengine.model import BaseModel
from mmengine.structures import BaseDataElement
from torch import nn
from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.models.mutators import OneShotModuleMutator
from mmrazor.registry import MODELS
from mmrazor.utils import SingleMutatorRandomSubnet, ValidFixMutable
from ..base import BaseAlgorithm, LossResults


@MODELS.register_module()
class SPOS(BaseAlgorithm):
    """Implementation of `SPOS <https://arxiv.org/abs/1904.00420>`_

    SPOS means Single Path One-Shot, a classic NAS algorithm.
    :class:`SPOS` implements the APIs required by the Single Path One-Shot
    algorithm, as well as the supernet training and subnet retraining logic
    for each iter.

    The logic of the search part is implemented in
    :class:`mmrazor.core.EvolutionSearch`

    Args:
        architecture (dict|:obj:`BaseModel`): The config of :class:`BaseModel`
            or built model. Corresponding to supernet in NAS algorithm.
        mutator (dict|:obj:`OneShotModuleMutator`): The config of
            :class:`OneShotModuleMutator` or built mutator.
        fix_subnet (str | dict | :obj:`FixSubnet`): The path of yaml file or
            loaded dict or built :obj:`FixSubnet`.
        norm_training (bool): Whether to set norm layers to training mode,
            namely, not freeze running stats (mean and var). Note: Effect on
            Batch Norm and its variants only. Defaults to False.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`. Defaults to None.
        init_cfg (dict): Init config for ``BaseModule``.

    Note:
        SPOS has two training mode: supernet training and subnet retraining.
        If `fix_subnet` is None, it means supernet training.
        If `fix_subnet` is not None, it means subnet training.

    Note:
        During supernet training, since each op is not fully trained, the
        statistics of :obj:_BatchNorm are inaccurate. This problem affects the
        evaluation of the performance of each subnet in the search phase. There
        are usually two ways to solve this problem, both need to set
        `norm_training` to True:

        1) Using a large batch size, BNs use the mean and variance of the
           current batch during forward.
        2) Recalibrate the statistics of BN before searching.

    Note:
        SPOS only uses one mutator. If you want to inherit SPOS to develop
        more complex algorithms, it is also feasible to use multiple mutators.
        For example, one part of the supernet uses SPOS(OneShotModuleMutator)
        to search, and the other part uses Darts(DiffModuleMutator) to search.
    """

    # TODO fix ea's name in doc-string.

    def __init__(self,
                 architecture: Union[BaseModel, Dict],
                 mutator: Optional[Union[OneShotModuleMutator, Dict]] = None,
                 fix_subnet: Optional[ValidFixMutable] = None,
                 norm_training: bool = False,
                 data_preprocessor: Optional[Union[dict, nn.Module]] = None,
                 init_cfg: Optional[dict] = None):
        super().__init__(architecture, data_preprocessor, init_cfg)

        # SPOS has two training mode: supernet training and subnet retraining.
        # fix_subnet is not None, means subnet retraining.
        if fix_subnet:
            # Avoid circular import
            from mmrazor.structures import load_fix_subnet

            # According to fix_subnet, delete the unchosen part of supernet
            load_fix_subnet(self.architecture, fix_subnet)
            self.is_supernet = False
        else:
            assert mutator is not None, \
                'mutator cannot be None when fix_subnet is None.'
            if isinstance(mutator, OneShotModuleMutator):
                self.mutator = mutator
            elif isinstance(mutator, dict):
                self.mutator = MODELS.build(mutator)
            else:
                raise TypeError('mutator should be a `dict` or '
                                f'`OneShotModuleMutator` instance, but got '
                                f'{type(mutator)}')

            # Mutator is an essential component of the NAS algorithm. It
            # provides some APIs commonly used by NAS.
            # Before using it, you must do some preparations according to
            # the supernet.
            self.mutator.prepare_from_supernet(self.architecture)
            self.is_supernet = True

        self.norm_training = norm_training

    def sample_subnet(self) -> SingleMutatorRandomSubnet:
        """Random sample subnet by mutator."""
        return self.mutator.sample_choices()

    def set_subnet(self, subnet: SingleMutatorRandomSubnet):
        """Set the subnet sampled by :meth:sample_subnet."""
        self.mutator.set_choices(subnet)

    def loss(
        self,
        batch_inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
    ) -> LossResults:
        """Calculate losses from a batch of inputs and data samples."""
        if self.is_supernet:
            random_subnet = self.sample_subnet()
            self.set_subnet(random_subnet)
            return self.architecture(batch_inputs, data_samples, mode='loss')
        else:
            return self.architecture(batch_inputs, data_samples, mode='loss')

    def train(self, mode=True):
        """Convert the model into eval mode while keep normalization layer
        unfreezed."""

        super().train(mode)
        if self.norm_training and not mode:
            for module in self.architecture.modules():
                if isinstance(module, _BatchNorm):
                    module.training = True



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABC, abstractmethod
from typing import Any, Optional, Set

from torch import nn

from mmrazor.models.mutables.base_mutable import BaseMutable


class DynamicOP(ABC):
    """Base class for dynamic OP. A dynamic OP usually consists of a normal
    static OP and mutables, where mutables are used to control the searchable
    (mutable) part of the dynamic OP.

    Note:
        When the dynamic OP has just been initialized, its forward propagation
        logic should be the same as the corresponding static OP. Only after
        the searchable part accepts the specific mutable through the
        corresponding interface does the part really become dynamic.

    Note:
        All subclass should implement ``to_static_op`` API.

    Args:
        accepted_mutables (set): The string set of all accepted mutables.
    """
    accepted_mutables: Set[str] = set()

    @abstractmethod
    def to_static_op(self) -> nn.Module:
        """Convert dynamic OP to static OP.

        Note:
            The forward result for the same input between dynamic OP and its
            corresponding static OP must be same.

        Returns:
            nn.Module: Corresponding static OP.
        """

    def check_if_mutables_fixed(self) -> None:
        """Check if all mutables are fixed.

        Raises:
            RuntimeError: Error if a existing mutable is not fixed.
        """

        def check_fixed(mutable: Optional[BaseMutable]) -> None:
            if mutable is not None and not mutable.is_fixed:
                raise RuntimeError(f'Mutable {type(mutable)} is not fixed.')

        for mutable in self.accepted_mutables:
            check_fixed(getattr(self, f'{mutable}'))

    @staticmethod
    def get_current_choice(mutable: BaseMutable) -> Any:
        """Get current choice of given mutable.

        Args:
            mutable (BaseMutable): Given mutable.

        Raises:
            RuntimeError: Error if `current_choice` is None.

        Returns:
            Any: Current choice of given mutable.
        """
        current_choice = mutable.current_choice
        if current_choice is None:
            raise RuntimeError(f'current choice of mutable {type(mutable)} '
                               'can not be None at runtime')

        return current_choice


class ChannelDynamicOP(DynamicOP):
    """Base class for dynamic OP with mutable channels.

    Note:
        All subclass should implement ``mutable_in`` and ``mutable_out`` APIs.
    """

    @property
    @abstractmethod
    def mutable_in(self) -> Optional[BaseMutable]:
        """Mutable related to input."""

    @property
    @abstractmethod
    def mutable_out(self) -> Optional[BaseMutable]:
        """Mutable related to output."""

    @staticmethod
    def check_mutable_channels(mutable_channels: BaseMutable) -> None:
        """Check if mutable has `currnet_mask` attribute.

        Args:
            mutable_channels (BaseMutable): Mutable to be checked.

        Raises:
            ValueError: Error if mutable does not have `current_mask`
                attribute.
        """
        if not hasattr(mutable_channels, 'current_mask'):
            raise ValueError(
                'channel mutable must have attribute `current_mask`')



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np

from mmrazor.registry import TASK_UTILS
from ..flops_params_counter import get_model_parameters_number
from .base_counter import BaseCounter


@TASK_UTILS.register_module()
class LinearCounter(BaseCounter):
    """FLOPs/params counter for Linear operation series."""

    @staticmethod
    def add_count_hook(module, input, output):
        input = input[0]
        output_last_dim = output.shape[
            -1]  # pytorch checks dimensions, so here we don't care much
        module.__flops__ += int(np.prod(input.shape) * output_last_dim)
        module.__params__ += get_model_parameters_number(module)
